from .utils import ig, ig_net

def explain(args, model, dataset, feature_names, omic_group, labels, name=None):
    print('Perform model interpretation')
    if args.model == "DNN" or name == "DNN":
        for target in set(labels):
            ig(args, model, dataset, feature_names, omic_group, target=int(target))
    if args.model == "Net" or name == "Net":
        for target in set(labels):
            ig_net(args, model, dataset, feature_names, omic_group, target=int(target))





import os
from .utils import check_files, init_log
from .fsd import feature_selection_distribution
from .train import train, ml_models
from .train import train_net
from .train import train_moanna, train_mogonet
from .explain import explain


def run(args):
    # check files exists
    check_files(args.omic_file)
    check_files(args.label_file)

    # make documents
    if not os.path.exists(args.outdir):
        os.makedirs(args.outdir)

    # init log.txt
    init_log(args=args)

    # features selection - with FSD or not
    data, feature_name, feature_group, labels = feature_selection_distribution(args)

    # training model - using different models
    if args.model == "DNN":
        model, dataset_test = train(args, data, labels)
        # # explain model
        # explain(args, model, dataset_test, feature_name, feature_group, labels)
    elif args.model in ["RF", "XGboost", "svm"]:
        name = args.model
        ml_models(args, data, feature_name, feature_group, labels, model_name=name)
    elif args.model == "Net":
        model, dataset_test = train_net(args, data, feature_group, labels)
        # # explain model
        # explain(args, model, dataset_test, feature_name, feature_group, labels)
    
    elif args.model == "moanna":
        train_moanna(args, data, labels)

    elif args.model == "mogonet":
        train_mogonet(args, data, labels)
    
    # multiple models
    elif args.model == "all":
        # DNN model
        model, dataset_test = train(args, data, labels)
        # explain(args, model, dataset_test, feature_name, feature_group, labels, name="DNN")
        # Net model
        model, dataset_test = train_net(args, data, feature_group, labels)
        # explain(args, model, dataset_test, feature_name, feature_group, labels, name="Net")
        # ml model
        for name in ["RF", "XGboost", "svm"]:
            ml_models(args, data, feature_name, feature_group, labels, model_name=name)














import os, sys
import numpy as np
import pandas as pd 
import torch
from pathlib import Path

from model.Moanna import Moanna
from helper.DataCleanUpFunctions import convert_dat_df_to_tensor, convert_list_labels_to_tensor

def moanna_predict(input, out_summary, out_details, model_state, model_json):

    base_path = Path(__file__).parent
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    # Load Model
    moanna = load_model(model_state, model_json)
    moanna.eval()

    # Read Input
    dat_df = pd.read_csv(input, sep="\t", index_col=0)

    # Process Input
    dat_tensor = convert_dat_df_to_tensor(dat_df)
    dat_tensor = dat_tensor.to(device)

    # Run Moanna
    encoded, decoded, predictions = moanna(dat_tensor)

    # Get Output
    ## TODO: Parameterised
    type_of_predictions={0:"ERStatus", 1:"HER2Status", 2: "Pam50Subtype"}

    pred_class_dict = {}
    pred_percent_list = []

    for i, label in type_of_predictions.items():
        # Get prediction labels and its string mapping
        mapping_file = os.path.join(base_path, "mapping_file", label+".txt")
        labels_map   = get_label_mapping(label, mapping_file)
        all_labels   = list(labels_map.values())

        # Get prediction 
        pred_classes = get_pred_classes(predictions[i])
        pred_percent_list.append(pd.DataFrame(get_pred_percentage(predictions[i]), index=dat_df.index.to_list(), columns=all_labels))

        # Map prediction label to string
        pred_class_dict["Pred_" + label] = [labels_map [x] for x in pred_classes]

    pred_class_df = pd.DataFrame(pred_class_dict, index=dat_df.index.to_list())
    pred_percent_df = pd.concat(pred_percent_list, axis=1, sort=False)

    # Write output
    pred_class_df.to_csv(out_summary, sep="\t")
    pred_percent_df.to_csv(out_details, sep="\t")

    return 0

def moanna_train(input=None, label=None, output='test'):

    # Basic setup
    base_path = Path(__file__).parent
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    # Read Inputs & Labels 
    # dat_df = pd.read_csv(input, sep="\t", index_col=0)
    df1 = pd.read_csv(r'C:\Users\liangbilin\Desktop\AttentionMOI\dataset\BLCA_select_data\BLCA_cnv.csv', index_col=0)
    df2 = pd.read_csv(r'C:\Users\liangbilin\Desktop\AttentionMOI\dataset\BLCA_select_data\BLCA_met.csv', index_col=0)
    df3 = pd.read_csv(r'C:\Users\liangbilin\Desktop\AttentionMOI\dataset\BLCA_select_data\BLCA_rna.csv', index_col=0)   
    patients = df1.index.values.tolist() + df2.index.values.tolist() + df3.index.values.tolist()
    patients_count = pd.DataFrame(pd.value_counts(patients))
    patients_extracted = patients_count[patients_count.iloc[:, 0]==3].index.values
    dat_df = pd.concat( [df1.loc[patients_extracted,:], df2.loc[patients_extracted,:], df3.loc[patients_extracted,:]], axis=1)

    # labels_df = pd.read_csv(label, sep="\t", index_col=0)
    # list_label_df = [labels_df[col] for col in labels_df.columns.tolist()]
    list_label_df = pd.read_csv(r'C:\Users\liangbilin\Desktop\AttentionMOI\dataset\BLCA_select_data\BLCA_label.csv').label.values.tolist()

    # Pre-defined hyperparameters
    ## TODO: Parameterised 
    params = {}
    params["input_size"] = dat_df.shape[1] # Number of features
    params["n_layers"] = 1
    params["encoded_size"] = 64 # Number of Autoencoders neurons
    params["hidden_size"] = params["encoded_size"] * (2 ** (params["n_layers"] + 1))
    params["batch_size"] = dat_df.shape[0] # Number of samples
    params["drop_prob"] = 0.5
    params["fnn_hidden_size"] = 40
    params["fnn_number_layers"] = 2
    params["num_classes"] = [len(np.unique(classes)) for classes in list_label_df]
    params["fnn_num_epoch"] = 100
    params["fnn_learning_rate"] = 0.05

    # Setup model
    moanna = Moanna(
        params["input_size"], 
        params["hidden_size"], 
        params["encoded_size"], 
        params["n_layers"], 
        params["drop_prob"], 
        params["fnn_hidden_size"], 
        params["num_classes"], 
        params["fnn_number_layers"], 
        0.1,
    )
    moanna = moanna.to(device)

    # Training optimiser
    moanna_optimizer = torch.optim.SGD(moanna.parameters(), lr=params["fnn_learning_rate"] , momentum=0.9)

    # Loss functions
    criterion1 = torch.nn.MSELoss(reduce=True)
    classifiers_criterion = []
    for classes in params["num_classes"]:
        classifiers_criterion.append(torch.nn.CrossEntropyLoss())   

    # Prep inputs
    dat_tensor = convert_dat_df_to_tensor(dat_df)
    # list_label_tensor = convert_list_labels_to_tensor(list_label_df) 
    list_label_tensor = torch.Tensor(list_label_df).reshape(-1,1)

    # For GPU's training
    dat_tensor = dat_tensor.to(device)
    for i in range(len(list_label_tensor)):
        list_label_tensor[i] = list_label_tensor[i].to(device)   

    # Train model
    for epoch in range(params["fnn_num_epoch"]):
        moanna_optimizer.zero_grad()
        encoded, decoded, predictions = moanna(dat_tensor)
    
        losses = [criterion1(decoded, dat_tensor)]
        for i in range(len(predictions)):
            print(len(predictions), list_label_tensor[i])
            losses.append(classifiers_criterion[i](predictions[i], list_label_tensor[i]))
        loss  = sum(losses)
    
        print ("Epoch %s: loss=%s" %(epoch,loss))
    
        loss.backward()
        moanna_optimizer.step()

    # Save model
    save_model(moanna, output, params)
    
    return 0

def get_pred_percentage(pred):
    val = torch.softmax(pred, dim=1)
    val_decimal = val.cpu().detach().numpy()
    val_percent = pd.DataFrame(val_decimal).applymap(lambda x: '{:.2%}'.format(x)).values
    return val_percent
    
def get_pred_classes(pred):
    val = torch.softmax(pred, dim=1)
    _, _classes = torch.max(val, 1)  
    classes = _classes.cpu().numpy()
    return classes

def get_pred_max(pred):
    val = torch.softmax(pred, dim=1)
    val_decimal = val.cpu().detach().numpy()  
    val_percent = pd.DataFrame(val_decimal.max(axis=1)).applymap(lambda x: '{:.2%}'.format(x)).values
    return val_percent

def get_label_mapping(column, mapping_file):
    '''
    PyTorch label is numeric and this function maps it back to a string label
    '''
    open_map_file = open(mapping_file)

    subtype_labels = {}

    for eachrow in open_map_file:
        eachrow = eachrow.strip().split("\t")
        file_column_name = eachrow[0]
        file_column_key  = eachrow[1]
        try:
            file_column_value= int(eachrow[2])
            file_column_label= eachrow[3]
        except:
            print ("map_file third column has to be an integer")
            sys.exit(1)

        if file_column_name == column:
            subtype_labels.update({file_column_value:file_column_label})

    return subtype_labels

def save_model(model, output, params):
    import json

    torch.save(model.state_dict(), output+".pt")

    with open(output+".json", 'w') as file:
        json.dump(params, file)

def load_model(model_state, model_json):
    import json

    with open(model_json, 'r') as file:
        params = json.load(file)      

    model = Moanna(
        params["input_size"], 
        params["hidden_size"], 
        params["encoded_size"], 
        params["n_layers"], 
        params["drop_prob"], 
        params["fnn_hidden_size"], 
        params["num_classes"], 
        params["fnn_number_layers"], 
        0.1,
    )

    model.load_state_dict(torch.load(model_state))

    return model



if __name__ == '__main__':
    moanna_train()



import pandas as pd
from sklearn.feature_selection import SelectPercentile, RFE
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegressionCV
from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA


# ANOVA
def anova(args, df_omic, label):
    """
    args: parameters from deepmoi.py.
    df_omic: omic data with DataFrame format.
    label: sample label.

    Return:
        selected omic data with DataFrame format.
    """
    select = SelectPercentile(percentile=args.percentile)
    select.fit(X=df_omic.values, y=label)
    features = df_omic.columns.values[select.get_support()]  # features selected by ANOVA.
    return df_omic[features]

# Recursive Feature Elimination (RFE)
def rfe(args, df_omic, label):
    """
    args: parameters from deepmoi.py.
    df_omic: omic data with DataFrame format.
    label: sample label.

    Return:
        selected omic data with DataFrame format.
    """
    keep_feat_num = df_omic.shape[1] // 10  # 默认保留10%的特征
    select = RFE(estimator=RandomForestClassifier(n_estimators=50, random_state=args.seed),
                 n_features_to_select=keep_feat_num, step=0.01)
    select.fit(X=df_omic.values, y=label)
    features = df_omic.columns.values[select.get_support()]  # features selected by REF.
    return df_omic[features]


# LASSO
def lasso(df_omic, label):
    """
    args: parameters from deepmoi.py.
    df_omic: omic data with DataFrame format.
    label: sample label.

    Return:
        selected omic data with DataFrame format.
    """
    cs = [1e-4, 1e-3, 1e-2, 1e-1, 1e0, 1e1, 1e2, 1e3, 1e4]
    select = LogisticRegressionCV(Cs=cs, cv=5, penalty='l1', max_iter=1000, solver='liblinear')
    select.fit(X=df_omic.values, y=label)
    features = df_omic.columns[select.coef_[0] != 0]
    return df_omic[features]


# PCA
def pca(args, df_omic):
    """
    args: parameters from deepmoi.py.
    df_omic: omic data with DataFrame format.
    label: sample label.

    Return:
        selected omic data with DataFrame format.
    """
    num_pc = args.num_pc
    features = df_omic.columns
    x_stand = StandardScaler().fit(df_omic).transform(df_omic)
    pca_transformer = PCA().fit(x_stand)
    x_pc = pca_transformer.transform(x_stand)
    var_explained = pca_transformer.explained_variance_ratio_
    print("Variance explained by the first 2 PCs is {}, \nVariance explained by the first 30 PCs is {}.".format(sum(var_explained[0:2]), sum(var_explained[0:30])))
    print("Variance explained by {} PCs is {}.".format(num_pc, sum(var_explained[0:num_pc])))
    df_omic = pd.DataFrame(x_pc[:, 0:num_pc])
    df_omic.columns = ["PC_" + str(i + 1) for i in range(len(df_omic.columns))]
    f_contribution = pca_transformer.components_[0:num_pc, :].sum(axis=0)
    feature_imp = pd.DataFrame(data=f_contribution, index=features, columns=["feature_imp"])
    return df_omic, feature_imp



import sys
import argparse

class DefaultHelpArgParser(argparse.ArgumentParser):
    def error(self, message):
        sys.stderr.write("Error: %s\n" %message)
        self.print_help()
        sys.exit(1)

def process_args(sysargs=None):
    cmds = {
        "version": do_version,
        "predict": do_predict,
        "train": do_train,
    }

    parser = DefaultHelpArgParser(description="Run Moanna")

    parser.add_argument("-v", "--version", action="store_true")

    subparsers = parser.add_subparsers(dest="command")

    add_predict_args(subparsers.add_parser("predict", help="Run Moanna Pre-trained prediction"))
    add_train_args(subparsers.add_parser("train", help="Run Moanna training"))

    args = parser.parse_args(sysargs)
    
    if args.version:
        return do_version(args)

    if not args.command:
        parser.print_help()
        sys.exit(1)

    return cmds[args.command](args)

def do_version(args):
    from moanna.__meta__ import __version__
    print (__version__)

def do_predict(args):
    from moanna.main import moanna_predict

    proc = moanna_predict(
        args.input,
        args.summary,
        args.output,
        args.model,
        args.json,
    )

    if not (proc):
        print ("Moanna completed successfully")
    else:
        print ("Moanna completed with errors")
        
    raise SystemExit

def do_train(args):
    from moanna.main import moanna_train

    proc = moanna_train(
        args.input,
        args.label,
        args.output,
    )

    print (proc)

    raise SystemExit

def add_predict_args(parser, add_argument=True):

    parser.add_argument(
        "-i",
        "--input",
        help="Input features table",
        required=True,
    )

    parser.add_argument(
        "-s",
        "--summary",
        help="Output predictions summary table",
        required=True,
    )

    parser.add_argument(
        "-o",
        "--output",
        help="Output prediction detailed table",
        required=True,
    )

    parser.add_argument(
        "-m",
        "--model",
        help="Pre-trained Model PyTorch State",
        required=True,
    )

    parser.add_argument(
        "-j",
        "--json",
        help="Pre-trained Model JSON parameters",
        required=True,
    )

    return parser

def add_train_args(parser, add_argument=True):

    parser.add_argument(
        "-i",
        "--input",
        help="Input features table",
        required=True,
    )

    parser.add_argument(
        "-l",
        "--label",
        help="Labels for training",
        required=True,
    )

    parser.add_argument(
        "-o",
        "--output",
        help="Output table",
        required=True,
    )

    return parser


# Libraries
import torch
import torch.nn as nn
import numpy as np
import pandas as pd
import torch.nn.functional as F
import seaborn as sns
import matplotlib.pyplot as plt

from livelossplot import PlotLosses

#Encoder
class LayerBlockEncode(nn.Module):
    def __init__(self, hidden_size_1, hidden_size_2, p):
        super().__init__()
        self.layer = nn.Linear(hidden_size_1, hidden_size_2)
        self.activation = nn.Tanh()
        self.bn = nn.BatchNorm1d(hidden_size_2)
        self.dropout = nn.Dropout(p)
        
    def forward(self, x):
        x = self.layer(x)
        x = self.activation(x)
        x = self.bn(x)
        x = self.dropout(x)
        return (x)

class Encoder(nn.Module):
    def __init__(self, input_shape, hidden_size, encoded_size, n_layers=3, drop_prob=0.5):
        super(Encoder, self).__init__()
        self.n_layers = n_layers
        self.drop_prob = drop_prob
        
        self.e1 = nn.Linear(input_shape, hidden_size)
        self.activation1 = nn.Tanh()
        self.bn1 = nn.BatchNorm1d(hidden_size)
        
        if (hidden_size // (2**n_layers)) > encoded_size:
            self.layers = nn.ModuleList([])
            for i in range(n_layers):
                self.layers.append(LayerBlockEncode(hidden_size//(2**i), hidden_size//(2**(i+1)), self.drop_prob))
        else:
            self.n_layers = 0
                                      
        self.e2 = nn.Linear((hidden_size//(2**n_layers)), encoded_size)
        
    def forward(self, input):
        x = self.e1(input)
        x = self.activation1(x)
        x = self.bn1(x)
        
        for i in range(self.n_layers):
            encode_block = self.layers[i]
            x = encode_block(x)
            
        #block1 = F.dropout(self.bn1(F.elu(self.e1(input))), p=self.drop_prob)
        #encoded_representation = torch.tanh(self.e2(block1))
        
        encoded_representation = self.e2(x)
        
        return encoded_representation
    
#Decoder:
class LayerBlockDecode(nn.Module):
    def __init__(self, hidden_size_1, hidden_size_2, p):
        super().__init__()
        self.layer = nn.Linear(hidden_size_1, hidden_size_2)
        self.activation = nn.Tanh()
        self.bn = nn.BatchNorm1d(hidden_size_2)
        self.dropout = nn.Dropout(p)
        
    def forward(self, x):
        x = self.layer(x)
        x = self.activation(x)
        x = self.bn(x)
        x = self.dropout(x)
        return (x)

class Decoder(nn.Module):
    def __init__(self, output_shape, hidden_size, encoded_size, n_layers=3, drop_prob=0.5):
        super(Decoder, self).__init__()
        self.n_layers = n_layers
        self.drop_prob = drop_prob
        self.second_last_layer_size = hidden_size // (2**n_layers)
        
        self.d1 = nn.Linear(encoded_size, self.second_last_layer_size)
        self.activation1 = nn.Tanh()
        self.bn1 = nn.BatchNorm1d(self.second_last_layer_size)
        
        if (self.second_last_layer_size) > encoded_size: 
            self.layers = nn.ModuleList([])
            for i in range(self.n_layers):
                self.layers.append(LayerBlockDecode(hidden_size//(2**(n_layers-i)), hidden_size//(2**(n_layers-i-1)), self.drop_prob))
        else:
            self.n_layers = 0
        
        self.d2 = nn.Linear(hidden_size, output_shape)
        
    
    def forward(self, input):
        x = self.d1(input)
        x = self.activation1(x)
        x = self.bn1(x)
        
        for i in range(self.n_layers):
            decode_block = self.layers[i]
            x = decode_block(x)
            
        #block = F.dropout(self.bn(F.elu(self.d(input))), p=self.drop_prob)
        #reconstruction = torch.tanh(self.d4(block))
        
        reconstruction = self.d2(x)
        
        return reconstruction
    
# Training AutoEncoders Function
def train_ae(input_tensor, target_tensor, encoder, decoder,
          encoder_optimizer, decoder_optimizer, criterion, hidden_size, encoded_size, n_layers, drop_prob, phase):
    
    if phase == 'train':
        encoder.train()
        decoder.train()
    else:
        encoder.eval()
        decoder.eval()
        
    # clear the gradients in the optimizers
    if phase == 'train':
        encoder_optimizer.zero_grad()
        decoder_optimizer.zero_grad()
    
    # Forward pass through 
    
    encoded_representation = encoder(input_tensor)
    reconstruction = decoder(encoded_representation)
    
    # Compute the loss
    loss = criterion(reconstruction, target_tensor)
    
    if phase == 'train':
        # Compute the gradients
        loss.backward()
    
        # Step the optimizers to update the model weights
        encoder_optimizer.step()
        decoder_optimizer.step()
    
    # Return the loss value to track training progress
    return loss.item()    

# Training Loop
def trainIters(encoder, decoder, data_tensor, data_tensor_valid, epochs, 
               hidden_size, encoded_size, n_layers, drop_prob,
               print_every_n_batches=100, learning_rate=0.01,
               phases=["train", "validation"],):
    
    # Live Loss
    liveloss = PlotLosses()
    
    # keep track of losses
    train_plot_losses = []
    test_plot_losses = []
    
    # Initialize Encoder Optimizer
    encoder_optimizer = torch.optim.Adam(encoder.parameters(), lr=learning_rate, weight_decay=1e-5)
    
    # Initialize Decoder Optimizer
    decoder_optimizer = torch.optim.Adam(decoder.parameters(), lr=learning_rate, weight_decay=1e-5)

    # Specify loss function
    criterion = torch.nn.MSELoss(reduce=True)
    
    # Cycle through epochs
    for epoch in range(epochs):
        logs = {}
        
        for phase in phases:
            print(f'Epoch {epoch + 1}/{epochs}')
            
            if phase == 'train':
                loss = train_ae(data_tensor, data_tensor, encoder, decoder,
                             encoder_optimizer, decoder_optimizer, criterion, 
                            hidden_size, encoded_size, n_layers, drop_prob, phase)
                train_plot_losses.append(loss)
            else:
                loss = train_ae(data_tensor_valid, data_tensor_valid, encoder, decoder,
                             encoder_optimizer, decoder_optimizer, criterion, 
                            hidden_size, encoded_size, n_layers, drop_prob, phase)
                test_plot_losses.append(loss)
                
            print(loss)
            #plot_losses.append(loss)
        
            prefix = ''
            if phase == 'validation':
                prefix = 'val_'
                
            logs[prefix + 'log loss'] = loss
        
        liveloss.update(logs) #liveloss
        liveloss.draw() #liveloss
                
    return train_plot_losses, test_plot_losses



import os
import numpy as np
import torch
import torch.nn.functional as F

cuda = True if torch.cuda.is_available() else False

def cal_sample_weight(labels, num_class, use_sample_weight=True):
    if not use_sample_weight:
        return np.ones(len(labels)) / len(labels)
    count = np.zeros(num_class)
    for i in range(num_class):
        count[i] = np.sum(labels==i)
    sample_weight = np.zeros(labels.shape)
    for i in range(num_class):
        sample_weight[np.where(labels==i)[0]] = count[i]/np.sum(count)
    
    return sample_weight


def one_hot_tensor(y, num_dim):
    y_onehot = torch.zeros(y.shape[0], num_dim)
    y_onehot.scatter_(1, y.view(-1,1), 1)
    
    return y_onehot


def cosine_distance_torch(x1, x2=None, eps=1e-8):
    x2 = x1 if x2 is None else x2
    w1 = x1.norm(p=2, dim=1, keepdim=True)
    w2 = w1 if x2 is x1 else x2.norm(p=2, dim=1, keepdim=True)
    return 1 - torch.mm(x1, x2.t()) / (w1 * w2.t()).clamp(min=eps)


def to_sparse(x):
    x_typename = torch.typename(x).split('.')[-1]
    sparse_tensortype = getattr(torch.sparse, x_typename)
    indices = torch.nonzero(x)
    if len(indices.shape) == 0:  # if all elements are zeros
        return sparse_tensortype(*x.shape)
    indices = indices.t()
    values = x[tuple(indices[i] for i in range(indices.shape[0]))]
    return sparse_tensortype(indices, values, x.size())


def cal_adj_mat_parameter(edge_per_node, data, metric="cosine"):
    assert metric == "cosine", "Only cosine distance implemented"
    dist = cosine_distance_torch(data, data)
    parameter = torch.sort(dist.reshape(-1,)).values[edge_per_node*data.shape[0]]
    return np.ndarray.item(parameter.data.cpu().numpy())


def graph_from_dist_tensor(dist, parameter, self_dist=True):
    if self_dist:
        assert dist.shape[0]==dist.shape[1], "Input is not pairwise dist matrix"
    g = (dist <= parameter).float()
    if self_dist:
        diag_idx = np.diag_indices(g.shape[0])
        g[diag_idx[0], diag_idx[1]] = 0
        
    return g


def gen_adj_mat_tensor(data, parameter, metric="cosine"):
    assert metric == "cosine", "Only cosine distance implemented"
    dist = cosine_distance_torch(data, data)
    g = graph_from_dist_tensor(dist, parameter, self_dist=True)
    if metric == "cosine":
        adj = 1-dist
    else:
        raise NotImplementedError
    adj = adj*g 
    adj_T = adj.transpose(0,1)
    I = torch.eye(adj.shape[0])
    if cuda:
        I = I.cuda()
    adj = adj + adj_T*(adj_T > adj).float() - adj*(adj_T > adj).float()
    adj = F.normalize(adj + I, p=1)
    adj = to_sparse(adj)
    
    return adj


def gen_test_adj_mat_tensor(data, trte_idx, parameter, metric="cosine"):
    assert metric == "cosine", "Only cosine distance implemented"
    adj = torch.zeros((data.shape[0], data.shape[0]))
    if cuda:
        adj = adj.cuda()
    num_tr = len(trte_idx["tr"])
    
    dist_tr2te = cosine_distance_torch(data[trte_idx["tr"]], data[trte_idx["te"]])
    g_tr2te = graph_from_dist_tensor(dist_tr2te, parameter, self_dist=False)
    if metric == "cosine":
        adj[:num_tr,num_tr:] = 1-dist_tr2te
    else:
        raise NotImplementedError
    adj[:num_tr,num_tr:] = adj[:num_tr,num_tr:]*g_tr2te
    
    dist_te2tr = cosine_distance_torch(data[trte_idx["te"]], data[trte_idx["tr"]])
    g_te2tr = graph_from_dist_tensor(dist_te2tr, parameter, self_dist=False)
    if metric == "cosine":
        adj[num_tr:,:num_tr] = 1-dist_te2tr
    else:
        raise NotImplementedError
    adj[num_tr:,:num_tr] = adj[num_tr:,:num_tr]*g_te2tr # retain selected edges
    
    adj_T = adj.transpose(0,1)
    I = torch.eye(adj.shape[0])
    if cuda:
        I = I.cuda()
    adj = adj + adj_T*(adj_T > adj).float() - adj*(adj_T > adj).float()
    adj = F.normalize(adj + I, p=1)
    adj = to_sparse(adj)
    
    return adj


def save_model_dict(folder, model_dict):
    if not os.path.exists(folder):
        os.makedirs(folder)
    for module in model_dict:
        torch.save(model_dict[module].state_dict(), os.path.join(folder, module+".pth"))
            
    
def load_model_dict(folder, model_dict):
    for module in model_dict:
        if os.path.exists(os.path.join(folder, module+".pth")):
#            print("Module {:} loaded!".format(module))
            model_dict[module].load_state_dict(torch.load(os.path.join(folder, module+".pth"), map_location="cuda:{:}".format(torch.cuda.current_device())))
        else:
            print("WARNING: Module {:} from model_dict is not loaded!".format(module))
        if cuda:
            model_dict[module].cuda()    
    return model_dict


import argparse, warnings
import numpy as np
from .src.main import run

warnings.filterwarnings('ignore')
np.random.seed(1234)


def get_args():
    parser = argparse.ArgumentParser()

    # config
    parser.add_argument('-f', '--omic_file', action='append', help='Input omics file.', required=True)
    parser.add_argument('-n', '--omic_name', action='append',
                        help='Input the omics name in the same order as the omics file.', required=True)
    parser.add_argument('-c', '--clin_file', default=None, help='Input clinical features file (optional)')
    parser.add_argument('-l', '--label_file', help='Input label file', required=True)
    parser.add_argument('-o', '--outdir', help='Setting output directory.', type=str, default='./output')

    # feature selection with distribution
    parser.add_argument('-i', '--iteration', help='The number of FSD iterations.', type=int, default=10)
    parser.add_argument('-s', '--seed', help='Default random number seed to use for FSD.', type=int, default=0)
    parser.add_argument('--threshold',
                        help='Threshold to use for FSD screening features,i.e. the proportion of iterations through FSD.',
                        type=float, default=0.8)

    # feature selection
    parser.add_argument('--method', help='Method of feature selection, choosing from ANOVA, RFE, LASSO, PCA.', type=str, default=None)
    parser.add_argument('--percentile', help='percentile for ANOVA.', type=float, default=30)
    parser.add_argument('--num_pc', help='number of PCs for PCA.', type=int, default=50)

    # whether using FSD
    parser.add_argument('--FSD', action="store_true", help='whether to use FSD to mitigate noise of omics. If set --FSD then using FSD, else not using FSD.')

    # building model
    parser.add_argument('-t', '--test_size', help='Testing dataset proportion.', type=float, default=0.3)
    parser.add_argument('-b', '--batch', help='Mini-batch number for model training.', type=int, default=32)
    parser.add_argument('-e', '--epoch', help='Epoch number for model training.', type=int, default=300)
    parser.add_argument('-r', '--lr', help='Learning rate for model training.', type=float, default=0.0001)
    parser.add_argument('-w', '--weight_decay', help='weight_decay parameter for model training.', type=float,
                        default=0.0001)

    # different models
    parser.add_argument('-m', '--model', help='Model names, choosing from RF, XGboost, SVM, DNN, MOGONET, Moanna, all', type=str, default="DNN")

    args = parser.parse_args()
    return args


def main():
    args = get_args()
    run(args)


if __name__ == "__main__":
    main()




def rename_feature(df, suffix):
    new_df_index = [x + suffix for x in df.index.tolist()]
    new_df_index_dict = {}
    for x in new_df_index:
        new_df_index_dict[x.split("_")[0]] = x
    df.rename(index=new_df_index_dict, inplace=True)

def standardise(df):
    df_transpose = df.transpose()
    df_transpose_rescale = (df_transpose - df_transpose.mean()) / df_transpose.std()
    return (df_transpose_rescale.transpose())

def print_shape(df, message, datatype):
    print(message, f'[{datatype}]: ', df.shape)

def print_mult_data_shape(expr_df, cnv_df, snp_df, message):
    print_shape(expr_df, message, "EXPR")
    print_shape(cnv_df, message, "CNV")
    print_shape(snp_df, message, "SNP")

def preprocess_tcga_df(expr_df, cnv_df, snp_df):
    print_mult_data_shape(expr_df, cnv_df, snp_df, "Original")
    
    # Re-header
    expr_df.set_index("Unnamed: 0", inplace=True)
    cnv_df.set_index("GeneName", inplace=True)
    snp_df.set_index("Unnamed: 0", inplace=True)
    
    # Replace "-" with "."
    cnv_df_list = list(cnv_df.columns.values)
    cnv_df_colnames = { x:x.replace("-",".") for x in cnv_df_list}
    cnv_df_tmp = cnv_df.rename(cnv_df_colnames, axis='columns')
    
    # Get the intersection of samplenames
    expr_df_list = list(expr_df.columns.values)
    cnv_df_list = list(cnv_df_tmp.columns.values)
    snp_df_list = list(snp_df.columns.values)

    samplenames = sorted(list(set(expr_df_list) & set(cnv_df_list) & set(snp_df_list)))

    expr_df_intersect = expr_df[expr_df.columns.intersection(samplenames)]
    cnv_df_intersect = cnv_df_tmp[cnv_df_tmp.columns.intersection(samplenames)]
    snp_df_intersect = snp_df[snp_df.columns.intersection(samplenames)]

    # Remove some rows from CNV, EXPR, SNP
    expr_df_rownames = expr_df_intersect.index.tolist()
    cnv_df_rownames = cnv_df_intersect.index.tolist()

    genenames = sorted(list(set(expr_df_rownames) & set(cnv_df_rownames)))

    cnv_df_final = cnv_df_intersect.loc[cnv_df_intersect.index.isin(genenames)]
    snp_df_final = snp_df_intersect
    expr_df_final = expr_df_intersect.loc[expr_df_intersect.index.isin(genenames)]
    expr_df_final.dropna(inplace=True)
    
    rename_feature(expr_df_final, "_EXPR")
    rename_feature(cnv_df_final, "_CNV")
    rename_feature(snp_df_final, "_SNP")
    
    print_mult_data_shape(expr_df_final, cnv_df_final, snp_df_final, "PreProcess")
    
    return expr_df_final, cnv_df_final, snp_df_final

def preprocess_metabric_df(expr_df, cnv_df, snp_df):
    print_mult_data_shape(expr_df, cnv_df, snp_df, "Original")
    
    # Re-header
    expr_df.set_index("Unnamed: 0", inplace=True)
    cnv_df.set_index("GeneName", inplace=True)
    snp_df.set_index("GeneName", inplace=True)
    
    # Replace "-" with "_"
    cnv_df_list = list(cnv_df.columns.values)
    cnv_df_colnames = { x:x.replace("-","_") for x in cnv_df_list}
    cnv_df_tmp = cnv_df.rename(cnv_df_colnames, axis='columns')
    
    # Replace "-" with "_"
    snp_df_list = list(snp_df.columns.values)
    snp_df_colnames = { x:x.replace("-","_") for x in snp_df_list}
    snp_df_tmp = snp_df.rename(snp_df_colnames, axis='columns')
    
    # Get the intersection of samplenames
    expr_df_list = list(expr_df.columns.values)
    cnv_df_list = list(cnv_df_tmp.columns.values)
    snp_df_list = list(snp_df_tmp.columns.values)

    samplenames = sorted(list(set(expr_df_list) & set(cnv_df_list) & set(snp_df_list)))

    expr_df_intersect = expr_df[expr_df.columns.intersection(samplenames)]
    cnv_df_intersect = cnv_df_tmp[cnv_df_tmp.columns.intersection(samplenames)]
    snp_df_intersect = snp_df_tmp[snp_df_tmp.columns.intersection(samplenames)]

    # Remove some rows from CNV
    expr_df_rownames = expr_df_intersect.index.tolist()
    cnv_df_rownames = cnv_df_intersect.index.tolist()

    genenames = sorted(list(set(expr_df_rownames) & set(cnv_df_rownames)))

    cnv_df_final = cnv_df_intersect.loc[cnv_df_intersect.index.isin(genenames)]
    snp_df_final = snp_df_intersect
    expr_df_final = expr_df_intersect.loc[expr_df_intersect.index.isin(genenames)]
    expr_df_final.dropna(inplace=True)
    
    rename_feature(expr_df_final, "_EXPR")
    rename_feature(cnv_df_final, "_CNV")
    rename_feature(snp_df_final, "_SNP")
    
    print_mult_data_shape(expr_df_final, cnv_df_final, snp_df_final, "PreProcess")
    
    return expr_df_final, cnv_df_final, snp_df_final

def preprocess_tcga_clin_df(clin_df_orig, expr_df):
    clin_df = clin_df_orig.copy()
    print_shape(clin_df, "Original", "ClinData")
    
    clin_df.set_index("SampleName", inplace=True)
    samplenames = clin_df.index.tolist()
    samplenames_expr = expr_df.columns.values
    samplenames_intersection = list(set(samplenames) & set(samplenames_expr))
    clin_df_final = clin_df.loc[clin_df.index.isin(samplenames_intersection)]
    clin_df_final = clin_df_final.reindex(expr_df.transpose().index)
                                    
    print_shape(clin_df_final, "PreProcess", "ClinData")
    
    return clin_df_final

def filter_clin_df(clin_df_orig, column, value):
    clin_df = clin_df_orig.copy()
    print_shape(clin_df, "Original", "ClinData")    
    
    # Filter
    clin_df_filtered = clin_df.loc[clin_df[column] == value]
    print_shape(clin_df_filtered, "Filtered", "ClinData")
    
    return clin_df_filtered

def add_tils_category(dat_df):
    import numpy as np
    
    new_dat_df = dat_df.copy()
    #Edit TILs info
    new_dat_df.loc[new_dat_df['tils_HandE_slides'].isna(),"TilsCategory"] = -1 # Will be replaced with NA at the last line
    new_dat_df.loc[(new_dat_df['tils_HandE_slides'] >= 0) & (new_dat_df['tils_HandE_slides'] < 30),"TilsCategory"] = 0
    new_dat_df.loc[new_dat_df['tils_HandE_slides'] >= 30,"TilsCategory"] = 1
    new_dat_df.loc[new_dat_df['tils_HandE_slides'] < 0,"TilsCategory"] = np.NaN
    
    return new_dat_df

def my_ttest(gene, group1, group2):
    import researchpy as rp
    descriptives, results = rp.ttest(group1[gene],group2[gene])
    ttest_pval = results.iloc[[3],[1]].values[0][0]
    
    return ttest_pval


def datasets_split(expr_df, cnv_df, snp_df, clin_df, split, method):
    if method == "random":
        split_method = split_samples_random
    
    train_samples, test_samples = split_method(clin_df, split)
    
    expr_df_train = expr_df[train_samples]
    expr_df_test = expr_df[test_samples]
    
    cnv_df_train = cnv_df[train_samples]
    cnv_df_test = cnv_df[test_samples]
    
    snp_df_train = snp_df[train_samples]
    snp_df_test = snp_df[test_samples]
    
    clin_df_train = clin_df.loc[train_samples]
    clin_df_test = clin_df.loc[test_samples]
    
    return expr_df_train, expr_df_test, cnv_df_train, cnv_df_test, snp_df_train, snp_df_test, clin_df_train, clin_df_test

def prep_data_label_remove_undefined(dat_tensor, label_df, label_type, labels_mapping):
    import pandas as pd
    import torch 
    import numpy as np
    
    # Get rid of Undefined
    if label_type == "Pam50Subtype":
        NAs = "Undefined"
    else:    
        NAs = "NA"
        
    undefined_list = label_df.index[label_df[label_type] == NAs].tolist()
    undefined_list += label_df.index[label_df[label_type].isna()].tolist()
    dat_df = pd.DataFrame(data = dat_tensor.detach().numpy(), index=label_df.index)
    
    dat_filtered_df = dat_df.loc[~dat_df.index.isin(undefined_list)]
    label_filtered_df = label_df.loc[~label_df.index.isin(undefined_list)].replace(labels_mapping)
    
    # convert to pytorch tensor
    dat_tensor = torch.from_numpy(dat_filtered_df.values).float()
    label_tensor = torch.from_numpy(label_filtered_df[label_type].values.astype(np.int32)).long().view(-1)
    
    return dat_tensor, label_tensor

def prep_data_classification_tcga(dat_tensor_train_orig, dat_tensor_valid_orig, label_df_train, label_df_valid, label_type, labels_mapping):
    
    dat_tensor_train, label_tensor_train = prep_data_label_remove_undefined(dat_tensor_train_orig, label_df_train, label_type, labels_mapping)
    dat_tensor_valid, label_tensor_valid = prep_data_label_remove_undefined(dat_tensor_valid_orig, label_df_valid, label_type, labels_mapping)
    
    return dat_tensor_train, dat_tensor_valid, label_tensor_train, label_tensor_valid
    
def standardise_and_merge(expr_df, cnv_df, snp_df):
    # Data Standardise + Merged 
    import pandas as pd
    
    expr_df_std = standardise(expr_df)

    frames=[expr_df_std, cnv_df, snp_df]
    data_df_orig = pd.concat(frames)
    data_df  = data_df_orig.transpose().copy()

    return data_df

def get_data_tensor(expr_df, cnv_df, snp_df):
    import torch

    data_df = standardise_and_merge(expr_df, cnv_df, snp_df)
    data_array = data_df.values
    data_tensor = torch.from_numpy(data_array).float()
    
    return data_tensor

def merge_data_3dfs(expr_df, cnv_df, snp_df):
    data_df = standardise_and_merge(expr_df, cnv_df, snp_df)
    return data_df

def merge_data_2dfs(expr_df, cnv_df):
    import pandas as pd
    
    expr_df_std = standardise(expr_df)
    frames=[expr_df_std, cnv_df]
    data_df_orig = pd.concat(frames)
    data_df  = data_df_orig.transpose().copy()
    
    return data_df

def convert_dat_df_to_tensor(data_df):
    import torch
    
    data_array = data_df.values
    data_tensor = torch.from_numpy(data_array).float()
    return data_tensor

def convert_label_df_to_tensor(label_df):
    import torch
    import numpy as np
    
    label_array  = label_df.values.astype(np.int32)
    label_tensor = torch.from_numpy(label_array).long().view(-1)
    return label_tensor

def convert_list_labels_to_tensor(list_label_df):
    return [convert_label_df_to_tensor(label_df) for label_df in list_label_df] 

def get_label_mapping(column):
    pam50_subtype_labels = {"Basal":0, "Her2":1, "LumA":2, "LumB":3, "Normal":4}
    tils_prediction = {1:1, 0:0}
    
    if column == "Pam50Subtype":
        subtype_labels = pam50_subtype_labels
    elif column == "tils_HandE_slides":
        subtype_labels = tils_prediction
    else:    
        subtype_labels = {"Positive":0, "Negative":1}
        
    return column, subtype_labels

def preprocess_two_datasets(tcga_df, metabric_df):
    
    tcga_df = tcga_df.loc[~tcga_df.index.duplicated(keep='first')]
    metabric_df = metabric_df.loc[~metabric_df.index.duplicated(keep='first')]
    
    tcga_df_rownames = tcga_df.index.tolist()
    metabric_df_rownames = metabric_df.index.tolist()
    genenames = sorted(list(set(tcga_df_rownames) & set(metabric_df_rownames)))
    
    tcga_df_final = tcga_df.loc[tcga_df.index.isin(genenames)].reindex(genenames)
    metabric_df_final = metabric_df.loc[metabric_df.index.isin(genenames)].reindex(genenames)
    
    return tcga_df_final, metabric_df_final

def get_label_mapping_file(column, map_file):
    import sys
    open_map_file = open(map_file)

    subtype_labels = {}

    for eachrow in open_map_file:
        eachrow = eachrow.strip().split("\t")
        file_column_name = eachrow[0]
        file_column_key  = eachrow[1]
        try:
            file_column_value= int(eachrow[2])
        except:
            print ("map_file third column has to be an integer")
            sys.exit(1)

        if file_column_name == column:
            subtype_labels.update({file_column_key:file_column_value})


    return column, subtype_labels

def rename_df_genenames(expr_df, cnv_df, snp_df):
    print_mult_data_shape(expr_df, cnv_df, snp_df, "Original")
    
    expr_df_final = expr_df.copy()
    cnv_df_final  = cnv_df.copy()
    snp_df_final  = snp_df.copy()
    
    rename_feature(expr_df_final, "_EXPR")
    rename_feature(cnv_df_final, "_CNV")
    rename_feature(snp_df_final, "_SNP")
    
    print_mult_data_shape(expr_df_final, cnv_df_final, snp_df_final, "PreProcess")
    
    return expr_df_final, cnv_df_final, snp_df_final  

def filter_clin_df_byColumn(clin_df, column):
    return clin_df[~clin_df[column].isna()]

def filter_clin_df_subtype(clin_df, subtype_list):
    return clin_df[clin_df["Pam50Subtype"].isin(subtype_list)]

def preprocess_dat_clin_df_byColumn(clin_df_orig, dat_df_orig, column):
    dat_df = dat_df_orig.transpose().copy()
    clin_df = clin_df_orig.copy()
    print_shape(clin_df, "Original", "ClinData")  
    clin_df.set_index("SampleName", inplace=True)
    
    clin_df = filter_clin_df_byColumn(clin_df, column)
    
    samplenames = clin_df.index.tolist()
    samplenames_dat = dat_df.columns.values
    samplenames_intersection = list(set(samplenames) & set(samplenames_dat))
    clin_df_final = clin_df.loc[clin_df.index.isin(samplenames_intersection)]
    dat_df_final  = dat_df[samplenames_intersection]
    clin_df_final = clin_df_final.reindex(dat_df_final.transpose().index)
    
    print_shape(clin_df_final, "PreProcess", "ClinData")
    print_shape(dat_df_final, "PreProcess", "MainData")
    
    return clin_df_final, dat_df_final.transpose()

def preprocess_dat_clin_df(clin_df_orig, dat_df_orig, subtype_list):
    dat_df = dat_df_orig.transpose().copy()
    clin_df = clin_df_orig.copy()
    print_shape(clin_df, "Original", "ClinData")  
    clin_df.set_index("SampleName", inplace=True)
    
    clin_df = filter_clin_df_subtype(clin_df, subtype_list)
    
    samplenames = clin_df.index.tolist()
    samplenames_dat = dat_df.columns.values
    samplenames_intersection = list(set(samplenames) & set(samplenames_dat))
    clin_df_final = clin_df.loc[clin_df.index.isin(samplenames_intersection)]
    dat_df_final  = dat_df[samplenames_intersection]
    clin_df_final = clin_df_final.reindex(dat_df_final.transpose().index)
    
    print_shape(clin_df_final, "PreProcess", "ClinData")
    print_shape(dat_df_final, "PreProcess", "MainData")
    
    return clin_df_final, dat_df_final.transpose()

def prep_data_label_remove_undefined2(dat_tensor_src, label_df_src, labels): #label_type, labels_mapping
    import pandas as pd
    import torch 
    import numpy as np
    
    dat_tensor_w = dat_tensor_src
    label_df_w = label_df_src.copy()
    
    list_labels = []
    list_labels_tensor = []
    
    print (labels)
    
    # Find all NAs / Undefined values that we need to clean
    # All samples ID collected to 'undefined_list'
    # Across multiple labels that we need to train (labels)
    undefined_list = []
    for label_id in labels:
        label_type = labels[label_id][0]
        
        # Get rid of Undefined
        if label_type == "Pam50Subtype":
            NAs = "Undefined"
        else:    
            NAs = "NA"
        
        undefined_list += label_df_w.index[label_df_w[label_type] == NAs].tolist()
        undefined_list += label_df_w.index[label_df_w[label_type].isna()].tolist()
        
        undefined_list = list(set(undefined_list))
        #print ("undefined_list", len(undefined_list), undefined_list)
        
    # Remove samples in 'undefined list' from data df
    # Making sure orders of indices are the same between data and labels 
    dat_df = pd.DataFrame(data = dat_tensor_w.detach().numpy(), index=label_df_w.index)    
    dat_filtered_df = dat_df.loc[~dat_df.index.isin(undefined_list)]
    #print ("dat_filtered_df", dat_filtered_df.shape, dat_filtered_df)
    
    # Remove samples in 'undefined list' from labels df too
    # Swap the mapping from characters to numeric (e.g. Positive -> 0; Negative->1)
    for label_id in labels:
        label_type = labels[label_id][0]
        labels_mapping = labels[label_id][1]
        label_filtered_df = label_df_w.loc[~label_df_w.index.isin(undefined_list)][label_type].replace(labels_mapping) 
        #print ("label_fitered_df", label_filtered_df.shape, label_filtered_df)
        list_labels.append(label_filtered_df)
    
    # convert to pytorch tensor
    dat_tensor = torch.from_numpy(dat_filtered_df.values).float() 
    print ("dat_tensor shape:", dat_tensor.shape)
    
    for label_filtered_df in list_labels:
        label_tensor = torch.from_numpy(label_filtered_df.values.astype(np.int32)).long().view(-1)
        list_labels_tensor.append(label_tensor)
        print("label_tensor shape:", label_tensor.shape)
        
    return dat_tensor, list_labels_tensor

def prep_data_classification2(dat_tensor_train_orig, dat_tensor_valid_orig, label_df_train, label_df_valid, labels):
    
    print (dat_tensor_train_orig.shape, label_df_train.shape)
    print (dat_tensor_valid_orig.shape, label_df_valid.shape)
    
    dat_tensor_train, list_label_tensor_train = prep_data_label_remove_undefined2(dat_tensor_train_orig, label_df_train, labels)
    dat_tensor_valid, list_label_tensor_valid = prep_data_label_remove_undefined2(dat_tensor_valid_orig, label_df_valid, labels)
    
    return dat_tensor_train, dat_tensor_valid, list_label_tensor_train, list_label_tensor_valid

def prep_data_classification_single(dat_tensor, label_df, alllabels):
    
    print (dat_tensor.shape, label_df.shape)
    
    dat_tensor, list_label_tensor = prep_data_label_remove_undefined2(dat_tensor, label_df, alllabels)
    
    return dat_tensor, list_label_tensor

def split_samples_balance(clin_df, column, split=0.7):
    import random

    all_samples = clin_df.index.tolist()
    sample_numbers = int(len(all_samples) * split)
    
    all_classes = list(set(clin_df[column].tolist()))
    min_samples = min(clin_df[column].value_counts())
    max_samples = max(clin_df[column].value_counts())
    sample_numbers_per_class = int(min_samples * split)
    
    train_samples = []
    for label_class in all_classes:
        class_sample = clin_df[clin_df[column] == label_class].index.tolist()
        train_class_sample = random.sample(class_sample, k=sample_numbers_per_class)
        train_samples += train_class_sample
        
    test_samples = [x for x in all_samples if x not in train_samples]
    
    return train_samples, test_samples

def split_samples_random(clin_df, column, split=0.7):
    import random
    
    all_samples = clin_df.index.tolist()
    sample_numbers = int(len(all_samples) * split)
    
    random.seed(12)
    train_samples = random.sample(all_samples, k=sample_numbers)
    test_samples = [x for x in all_samples if x not in train_samples]
    
    return train_samples, test_samples

def datasets_split2(dat_df, clin_df, column, split, method):
    if method == "random":
        split_method = split_samples_random
    elif method == "balance":
        split_method = split_samples_balance
    
    train_samples, test_samples = split_method(clin_df, column, split)
    
    dat_df = dat_df.copy().transpose()
    
    dat_df_train = dat_df[train_samples]
    dat_df_test = dat_df[test_samples]
    
    clin_df_train = clin_df.loc[train_samples]
    clin_df_test = clin_df.loc[test_samples]
    
    return dat_df_train.transpose(), dat_df_test.transpose(), clin_df_train, clin_df_test

def filter_df_for_NAs(dat_df_src, label_df_src, labels): #label_type, labels_mapping
    import pandas as pd
    import torch 
    import numpy as np
    
    print_shape(dat_df_src, "dat_df", "Prep Prior to Final Filtering")
    print_shape(label_df_src, "label_df", "Prep Prior to Final Filtering")
    
    dat_df = dat_df_src.copy()
    label_df_w = label_df_src.copy()
    
    list_labels = []
    list_labels_tensor = []
    
    print (labels)
    
    # Find all NAs / Undefined values that we need to clean
    # All samples ID collected to 'undefined_list'
    # Across multiple labels that we need to train (labels)
    undefined_list = []
    for label_id in labels:
        label_type = labels[label_id][0]
        
        # Get rid of Undefined
        if label_type == "Pam50Subtype":
            NAs = "Undefined"
        else:    
            NAs = "NA"
        
        undefined_list += label_df_w.index[label_df_w[label_type] == NAs].tolist()
        undefined_list += label_df_w.index[label_df_w[label_type].isna()].tolist()
        
        undefined_list = list(set(undefined_list))
        #print ("undefined_list", len(undefined_list), undefined_list)
        
    # Remove samples in 'undefined list' from data df
    # Making sure orders of indices are the same between data and labels 
    dat_df = dat_df.reindex(label_df_w.index)    
    dat_filtered_df = dat_df.loc[~dat_df.index.isin(undefined_list)]
    #print ("dat_filtered_df", dat_filtered_df.shape, dat_filtered_df)
    
    # Remove samples in 'undefined list' from labels df too
    # Swap the mapping from characters to numeric (e.g. Positive -> 0; Negative->1)
    for label_id in labels:
        label_type = labels[label_id][0]
        labels_mapping = labels[label_id][1]
        label_filtered_df = label_df_w.loc[~label_df_w.index.isin(undefined_list)][label_type].replace(labels_mapping) 
        #print ("label_fitered_df", label_filtered_df.shape, label_filtered_df)
        list_labels.append(label_filtered_df)
    
    print_shape(dat_filtered_df, "dat_df", "Post Filtering")
    for each_label_df in list_labels:
        print_shape(each_label_df, "label_df", "Post Filtering")
    
    return dat_filtered_df, list_labels


""" Training and testing of the model
"""
import os
import numpy as np
from sklearn.metrics import accuracy_score, f1_score, roc_auc_score
import torch
import torch.nn.functional as F
from .models import init_model_dict, init_optim
from .utils import one_hot_tensor, cal_sample_weight, gen_adj_mat_tensor, gen_test_adj_mat_tensor, cal_adj_mat_parameter

cuda = True if torch.cuda.is_available() else False


def prepare_trte_data(data_folder, view_list):
    num_view = len(view_list)
    labels_tr = np.loadtxt(os.path.join(data_folder, "labels_tr.csv"), delimiter=',')
    labels_te = np.loadtxt(os.path.join(data_folder, "labels_te.csv"), delimiter=',')
    labels_tr = labels_tr.astype(int)
    labels_te = labels_te.astype(int)
    data_tr_list = []
    data_te_list = []
    for i in view_list:
        data_tr_list.append(np.loadtxt(os.path.join(data_folder, str(i)+"_tr.csv"), delimiter=','))
        data_te_list.append(np.loadtxt(os.path.join(data_folder, str(i)+"_te.csv"), delimiter=','))
    num_tr = data_tr_list[0].shape[0]
    num_te = data_te_list[0].shape[0]
    data_mat_list = []
    for i in range(num_view):
        data_mat_list.append(np.concatenate((data_tr_list[i], data_te_list[i]), axis=0))
    data_tensor_list = []
    for i in range(len(data_mat_list)):
        data_tensor_list.append(torch.FloatTensor(data_mat_list[i]))
        if cuda:
            data_tensor_list[i] = data_tensor_list[i].cuda()
    idx_dict = {}
    idx_dict["tr"] = list(range(num_tr))
    idx_dict["te"] = list(range(num_tr, (num_tr+num_te)))
    data_train_list = []
    data_all_list = []
    for i in range(len(data_tensor_list)):
        data_train_list.append(data_tensor_list[i][idx_dict["tr"]].clone())
        data_all_list.append(torch.cat((data_tensor_list[i][idx_dict["tr"]].clone(),
                                       data_tensor_list[i][idx_dict["te"]].clone()),0))
    labels = np.concatenate((labels_tr, labels_te))
    
    return data_train_list, data_all_list, idx_dict, labels


def gen_trte_adj_mat(data_tr_list, data_trte_list, trte_idx, adj_parameter):
    adj_metric = "cosine" # cosine distance
    adj_train_list = []
    adj_test_list = []
    for i in range(len(data_tr_list)):
        adj_parameter_adaptive = cal_adj_mat_parameter(adj_parameter, data_tr_list[i], adj_metric)
        adj_train_list.append(gen_adj_mat_tensor(data_tr_list[i], adj_parameter_adaptive, adj_metric))
        adj_test_list.append(gen_test_adj_mat_tensor(data_trte_list[i], trte_idx, adj_parameter_adaptive, adj_metric))
    
    return adj_train_list, adj_test_list


def train_epoch(data_list, adj_list, label, one_hot_label, sample_weight, model_dict, optim_dict, train_VCDN=True):
    loss_dict = {}
    criterion = torch.nn.CrossEntropyLoss(reduction='none')
    for m in model_dict:
        model_dict[m].train()    
    num_view = len(data_list)
    for i in range(num_view):
        optim_dict["C{:}".format(i+1)].zero_grad()
        ci_loss = 0
        ci = model_dict["C{:}".format(i+1)](model_dict["E{:}".format(i+1)](data_list[i],adj_list[i]))
        ci_loss = torch.mean(torch.mul(criterion(ci, label),sample_weight))
        ci_loss.backward()
        optim_dict["C{:}".format(i+1)].step()
        loss_dict["C{:}".format(i+1)] = ci_loss.detach().cpu().numpy().item()
    if train_VCDN and num_view >= 2:
        optim_dict["C"].zero_grad()
        c_loss = 0
        ci_list = []
        for i in range(num_view):
            ci_list.append(model_dict["C{:}".format(i+1)](model_dict["E{:}".format(i+1)](data_list[i],adj_list[i])))
        c = model_dict["C"](ci_list)    
        c_loss = torch.mean(torch.mul(criterion(c, label),sample_weight))
        c_loss.backward()
        optim_dict["C"].step()
        loss_dict["C"] = c_loss.detach().cpu().numpy().item()
    
    return loss_dict
    

def test_epoch(data_list, adj_list, te_idx, model_dict):
    for m in model_dict:
        model_dict[m].eval()
    num_view = len(data_list)
    ci_list = []
    for i in range(num_view):
        ci_list.append(model_dict["C{:}".format(i+1)](model_dict["E{:}".format(i+1)](data_list[i],adj_list[i])))
    if num_view >= 2:
        c = model_dict["C"](ci_list)    
    else:
        c = ci_list[0]
    c = c[te_idx,:]
    prob = F.softmax(c, dim=1).data.cpu().numpy()
    
    return prob


# for evaluation
from sklearn.metrics import f1_score, recall_score, roc_auc_score, precision_score, accuracy_score
def evaluate(pred_prob, real_label, average="macro"):
    # For evaluating binary classification models
    if pred_prob.shape[1] == 2:
        y_pred = np.argmax(pred_prob, 1)
        prec = precision_score(real_label, y_pred)
        acc = accuracy_score(real_label, y_pred)
        f1 = f1_score(real_label, y_pred)
        recall = recall_score(real_label, y_pred)
        auc = roc_auc_score(real_label, pred_prob[:, 1])
    # For evaluating multiclass models
    else:
        y_pred = np.argmax(pred_prob, 1)
        prec = precision_score(real_label, y_pred, average='macro')
        acc = accuracy_score(real_label, y_pred)
        f1 = f1_score(real_label, y_pred, average=average)
        recall = recall_score(real_label, y_pred, average=average)
        auc = roc_auc_score(real_label, pred_prob, average='macro', multi_class='ovo')
    return acc, prec, f1, auc, recall

def evaluation(model, dataset):
    """
    To evaluate model performance for training dataset and testing dataset.

    model: taining model
    dataset: training dataset or testing dataset.
    """
    y_pred_probs, real_labels = [], []
    for d in dataset:
        out = model(d[0])
        y_pred_prob = F.softmax(out).detach().tolist()
        y_pred_probs.append(y_pred_prob)
        real_labels.append(d[1].tolist())
    acc, prec, f1, auc, recall = evaluate(pred_prob=np.array(y_pred_probs),
                                          real_label=np.array(real_labels))
    return acc, prec, f1, auc, recall


def train_test(data_folder, view_list, num_class,
               lr_e_pretrain, lr_e, lr_c, 
               num_epoch_pretrain, num_epoch):
    test_inverval = 50
    num_view = len(view_list)
    dim_hvcdn = pow(num_class,num_view)
    if data_folder == 'ROSMAP':
        adj_parameter = 2
        dim_he_list = [200,200,100]
    if data_folder == 'BRCA':
        adj_parameter = 10
        dim_he_list = [400,400,200]
    if data_folder == 'tmp':
        adj_parameter = 10
        dim_he_list = [200,200,100]
    data_tr_list, data_trte_list, trte_idx, labels_trte = prepare_trte_data(data_folder, view_list)
    labels_tr_tensor = torch.LongTensor(labels_trte[trte_idx["tr"]])
    onehot_labels_tr_tensor = one_hot_tensor(labels_tr_tensor, num_class)
    sample_weight_tr = cal_sample_weight(labels_trte[trte_idx["tr"]], num_class)
    sample_weight_tr = torch.FloatTensor(sample_weight_tr)
    if cuda:
        labels_tr_tensor = labels_tr_tensor.cuda()
        onehot_labels_tr_tensor = onehot_labels_tr_tensor.cuda()
        sample_weight_tr = sample_weight_tr.cuda()
    adj_tr_list, adj_te_list = gen_trte_adj_mat(data_tr_list, data_trte_list, trte_idx, adj_parameter)
    dim_list = [x.shape[1] for x in data_tr_list]
    model_dict = init_model_dict(num_view, num_class, dim_list, dim_he_list, dim_hvcdn)
    for m in model_dict:
        if cuda:
            model_dict[m].cuda()
    
    print("\nPretrain GCNs...")
    optim_dict = init_optim(num_view, model_dict, lr_e_pretrain, lr_c)
    for epoch in range(num_epoch_pretrain):
        train_epoch(data_tr_list, adj_tr_list, labels_tr_tensor, 
                    onehot_labels_tr_tensor, sample_weight_tr, model_dict, optim_dict, train_VCDN=False)
    print("\nTraining...")
    optim_dict = init_optim(num_view, model_dict, lr_e, lr_c)
    logs = []
    for epoch in range(num_epoch+1):
        train_epoch(data_tr_list, adj_tr_list, labels_tr_tensor, 
                    onehot_labels_tr_tensor, sample_weight_tr, model_dict, optim_dict)
        
        if epoch % test_inverval == 0:
            te_prob = test_epoch(data_trte_list, adj_te_list, trte_idx["te"], model_dict)
            # print("\nTest: Epoch {:d}".format(epoch))
            # if num_class == 2:
            #     print("Test ACC: {:.3f}".format(accuracy_score(labels_trte[trte_idx["te"]], te_prob.argmax(1))))
            #     print("Test F1: {:.3f}".format(f1_score(labels_trte[trte_idx["te"]], te_prob.argmax(1))))
            #     print("Test AUC: {:.3f}".format(roc_auc_score(labels_trte[trte_idx["te"]], te_prob[:,1])))
            # else:
            #     print("Test ACC: {:.3f}".format(accuracy_score(labels_trte[trte_idx["te"]], te_prob.argmax(1))))
            #     print("Test F1 weighted: {:.3f}".format(f1_score(labels_trte[trte_idx["te"]], te_prob.argmax(1), average='weighted')))
            #     print("Test F1 macro: {:.3f}".format(f1_score(labels_trte[trte_idx["te"]], te_prob.argmax(1), average='macro')))
        
        acc, prec, f1, auc, recall = evaluate(te_prob, labels_trte[trte_idx["te"]])
        loss_epoch = 0  # 占位
        log_test = 'Epoch {:2d} | Test_ACC {:.3f} | Test_AUC {:.3f} | Test_F1_score {:.3f} | Test_Recall {:.3f} | Test_Precision {:.3f}'.format(
            epoch, acc, auc, f1, recall, prec)
        print(log_test)  
        logs.append(log_test)
    
    return logs
        


""" Componets of the model
"""
import torch.nn as nn
import torch
import torch.nn.functional as F


def xavier_init(m):
    if type(m) == nn.Linear:
        nn.init.xavier_normal_(m.weight)
        if m.bias is not None:
           m.bias.data.fill_(0.0)
           

class GraphConvolution(nn.Module):
    def __init__(self, in_features, out_features, bias=True):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = nn.Parameter(torch.FloatTensor(in_features, out_features))
        if bias:
            self.bias = nn.Parameter(torch.FloatTensor(out_features))
        nn.init.xavier_normal_(self.weight.data)
        if self.bias is not None:
            self.bias.data.fill_(0.0)
    
    def forward(self, x, adj):
        support = torch.mm(x, self.weight)
        output = torch.sparse.mm(adj, support)
        if self.bias is not None:
            return output + self.bias
        else:
            return output
    

class GCN_E(nn.Module):
    def __init__(self, in_dim, hgcn_dim, dropout):
        super().__init__()
        self.gc1 = GraphConvolution(in_dim, hgcn_dim[0])
        self.gc2 = GraphConvolution(hgcn_dim[0], hgcn_dim[1])
        self.gc3 = GraphConvolution(hgcn_dim[1], hgcn_dim[2])
        self.dropout = dropout

    def forward(self, x, adj):
        x = self.gc1(x, adj)
        x = F.leaky_relu(x, 0.25)
        x = F.dropout(x, self.dropout, training=self.training)
        x = self.gc2(x, adj)
        x = F.leaky_relu(x, 0.25)
        x = F.dropout(x, self.dropout, training=self.training)
        x = self.gc3(x, adj)
        x = F.leaky_relu(x, 0.25)
        
        return x


class Classifier_1(nn.Module):
    def __init__(self, in_dim, out_dim):
        super().__init__()
        self.clf = nn.Sequential(nn.Linear(in_dim, out_dim))
        self.clf.apply(xavier_init)

    def forward(self, x):
        x = self.clf(x)
        return x


class VCDN(nn.Module):
    def __init__(self, num_view, num_cls, hvcdn_dim):
        super().__init__()
        self.num_cls = num_cls
        self.model = nn.Sequential(
            nn.Linear(pow(num_cls, num_view), hvcdn_dim),
            nn.LeakyReLU(0.25),
            nn.Linear(hvcdn_dim, num_cls)
        )
        self.model.apply(xavier_init)
        
    def forward(self, in_list):
        num_view = len(in_list)
        for i in range(num_view):
            in_list[i] = torch.sigmoid(in_list[i])
        x = torch.reshape(torch.matmul(in_list[0].unsqueeze(-1), in_list[1].unsqueeze(1)),(-1,pow(self.num_cls,2),1))
        for i in range(2,num_view):
            x = torch.reshape(torch.matmul(x, in_list[i].unsqueeze(1)),(-1,pow(self.num_cls,i+1),1))
        vcdn_feat = torch.reshape(x, (-1,pow(self.num_cls,num_view)))
        output = self.model(vcdn_feat)

        return output

    
def init_model_dict(num_view, num_class, dim_list, dim_he_list, dim_hc, gcn_dopout=0.5):
    model_dict = {}
    for i in range(num_view):
        model_dict["E{:}".format(i+1)] = GCN_E(dim_list[i], dim_he_list, gcn_dopout)
        model_dict["C{:}".format(i+1)] = Classifier_1(dim_he_list[-1], num_class)
    if num_view >= 2:
        model_dict["C"] = VCDN(num_view, num_class, dim_hc)
    return model_dict


def init_optim(num_view, model_dict, lr_e=1e-4, lr_c=1e-4):
    optim_dict = {}
    for i in range(num_view):
        optim_dict["C{:}".format(i+1)] = torch.optim.Adam(
                list(model_dict["E{:}".format(i+1)].parameters())+list(model_dict["C{:}".format(i+1)].parameters()), 
                lr=lr_e)
    if num_view >= 2:
        optim_dict["C"] = torch.optim.Adam(model_dict["C"].parameters(), lr=lr_c)
    return optim_dict


""" Example for MOGONET classification
"""
from .train_test import train_test


def run_mogonet(num_epoch=200):
    data_folder = 'tmp'
    view_list = [1,2,3]
    view_list = [1]
    num_epoch_pretrain = 500
    num_epoch = num_epoch
    lr_e_pretrain = 1e-3
    lr_e = 5e-4
    lr_c = 1e-3
    
    if data_folder == 'ROSMAP':
        num_class = 2
    if data_folder == 'BRCA':
        num_class = 5
    if data_folder == 'tmp':
        num_class = 3
    
    log = train_test(data_folder, view_list, num_class,
               lr_e_pretrain, lr_e, lr_c, 
               num_epoch_pretrain, num_epoch)
    
    return log


if __name__ == "__main__":    
    data_folder = 'tmp'
    view_list = [1,2,3]
    view_list = [1]
    num_epoch_pretrain = 500
    num_epoch = 2500
    lr_e_pretrain = 1e-3
    lr_e = 5e-4
    lr_c = 1e-3
    
    if data_folder == 'ROSMAP':
        num_class = 2
    if data_folder == 'BRCA':
        num_class = 5
    if data_folder == 'tmp':
        num_class = 3
    
    train_test(data_folder, view_list, num_class,
               lr_e_pretrain, lr_e, lr_c, 
               num_epoch_pretrain, num_epoch)             


import os
import torch
import numpy as np
import joblib
import torch.nn.functional as F
from .utils import check_files
from .preprocess import read_dataset, read_
from .utils import evaluate


def read_feature(args):
    if args.FSD:
        feature_file = os.path.join(args.outdir, 'features_FSD_{}_{}.csv'.format(args.method, args.seed))
    else:
        feature_file = os.path.join(args.outdir, 'features_{}_{}.csv'.format(args.method, args.seed))
    print("Featrue file path is: "+feature_file)
    features = read_(feature_file)
    return features

def predict_DNN(args, data):
    if args.FSD:
        if args.clin_file:
            # to load model
            model_path =  os.path.join(args.outdir,'model_DNN_FSD_{}_{}_clin_{}.pt'.format(args.method, args.omic_name, args.seed))
        else:
            # to load model
            model_path =  os.path.join(args.outdir,'model_DNN_FSD_{}_{}_{}.pt'.format(args.method, args.omic_name, args.seed))
    else:
        if args.clin_file:
            # to load model
            model_path =  os.path.join(args.outdir,'model_DNN_{}_{}_clin_{}.pt'.format(args.method, args.omic_name, args.seed))
        else:
            # to load model
            model_path =  os.path.join(args.outdir,'model_DNN_{}_{}_{}.pt'.format(args.method, args.omic_name, args.seed))
    print("Model path is: "+model_path)
    model = torch.load(model_path)
    out = model(torch.FloatTensor(data))
    out = out.squeeze(-1)
    y_pred_prob = F.softmax(out).detach().tolist()
    return model, y_pred_prob


def predict_Net(args, data, features):
    if args.FSD:
        if args.clin_file:
            # to load model
            model_path =  os.path.join(args.outdir,'model_Net_FSD_{}_{}_clin_{}.pt'.format(args.method, args.omic_name, args.seed))
        else:
            # to load model
            model_path =  os.path.join(args.outdir,'model_Net_FSD_{}_{}_{}.pt'.format(args.method, args.omic_name, args.seed))
    else:
        if args.clin_file:
            # to load model
            model_path =  os.path.join(args.outdir,'model_Net_{}_{}_clin_{}.pt'.format(args.method, args.omic_name, args.seed))
        else:
            # to load model
            model_path =  os.path.join(args.outdir,'model_Net_{}_{}_{}.pt'.format(args.method, args.omic_name, args.seed))
    print("Model path is: " + model_path)
    model = torch.load(model_path)
    idx_rna = []
    for i, v in enumerate(features["omic_group"]):
        if v == 'rna':
            idx_rna.append(i)

    idx_dna = []
    for i, v in enumerate(features["omic_group"]):
        if v == 'cnv' or v == 'met':
            idx_dna.append(i)

    data_dna = data[:, idx_dna]
    data_rna = data[:, idx_rna]
    out = model(torch.FloatTensor(data_dna), torch.FloatTensor(data_rna))
    out = out.squeeze(-1)
    y_pred_prob = F.softmax(out).detach().tolist()
    return model, y_pred_prob

def predict_ml(args, data, model_name):
    if args.FSD:
        if args.clin_file:
            # to load model
            model_path =  os.path.join(args.outdir,'model_{}_FSD_{}_{}_clin_{}.joblib'.format(model_name, args.method, args.omic_name, args.seed))
        else:
            # to load model
            model_path =  os.path.join(args.outdir,'model_{}_FSD_{}_{}_{}.joblib'.format(model_name, args.method, args.omic_name, args.seed))
    else:
        if args.clin_file:
            # to load model
            model_path =  os.path.join(args.outdir,'model_{}_{}_{}_clin_{}.joblib'.format(model_name, args.method, args.omic_name, args.seed))
        else:
            # to load model
            model_path =  os.path.join(args.outdir,'model_{}_{}_{}_{}.joblib'.format(model_name, args.method, args.omic_name, args.seed))
    print("Model path is: " + model_path)
    model = joblib.load(model_path)
    y_pred_prob = model.predict_proba(data)
    return model, y_pred_prob

def write_out(args, model_name, acc, prec, f1, auc, recall):
    with open(os.path.join(args.outdir, 'evaluation_external.txt'), 'a') as txt:
        if args.FSD:
            if args.clin_file:
                txt.writelines(
                    f"{model_name}\tFSD_{args.method}\t{args.omic_name} clin\t{acc}\t{prec}\t{f1}\t{auc}\t{recall}\n")
            else:
                txt.writelines(
                    f"{model_name}\tFSD_{args.method}\t{args.omic_name}\t{acc}\t{prec}\t{f1}\t{auc}\t{recall}\n")
        else:
            if args.clin_file:
                txt.writelines(
                    f"{model_name}\t{args.method}\t{args.omic_name} clin\t{acc}\t{prec}\t{f1}\t{auc}\t{recall}\n")
            else:
                txt.writelines(
                    f"{model_name}\t{args.method}\t{args.omic_name}\t{acc}\t{prec}\t{f1}\t{auc}\t{recall}\n")
    txt.close()

def predict(args):
    # check files exists
    check_files(args.omic_file)
    check_files(args.label_file)

    # make documents
    if not os.path.exists(args.outdir):
        os.makedirs(args.outdir)

    # read external datasets
    df_omics, df_label, df_clin = read_dataset(args)
    labels = df_label.label.values

    # select features
    features = read_feature(args)
    chosen_omics = []

    for omic_num, df_omic in enumerate(df_omics):
        feat_name = features[features["omic_group"] == args.omic_name[omic_num]]["feat_name"]
        df_omic = df_omic[feat_name]
        chosen_omics.append(df_omic.values)
    data = np.concatenate(chosen_omics, 1)
    data = np.nan_to_num(data)
    print("The shape of external dataset is {}.".format(data.shape))

    if args.model == "DNN":
        model, y_pred_prob = predict_DNN(args, data)
        # # evaluate model
        acc, prec, f1, auc, recall = evaluate(pred_prob=np.array(y_pred_prob), real_label=np.array(labels))
        write_out(args, "DNN", acc, prec, f1, auc, recall)
    elif args.model in ["RF", "XGboost", "svm"]:
        model_name = args.model
        model, y_pred_prob = predict_ml(args, data, model_name)
        # # evaluate model
        acc, prec, f1, auc, recall = evaluate(pred_prob=np.array(y_pred_prob), real_label=np.array(labels))
        write_out(args, model_name, acc, prec, f1, auc, recall)
    elif args.model == "Net":
        model, y_pred_prob = predict_Net(args, data, features)
        # # evaluate model
        acc, prec, f1, auc, recall = evaluate(pred_prob=np.array(y_pred_prob), real_label=np.array(labels))
        write_out(args, "Met", acc, prec, f1, auc, recall)
    # multiple models
    elif args.model == "all":
        # DNN model
        model, y_pred_prob = predict_DNN(args, data)
        # # evaluate model
        acc, prec, f1, auc, recall = evaluate(pred_prob=np.array(y_pred_prob), real_label=np.array(labels))
        write_out(args, "DNN", acc, prec, f1, auc, recall)
        # Net model
        model, y_pred_prob = predict_Net(args, data, features)
        # # evaluate model
        acc, prec, f1, auc, recall = evaluate(pred_prob=np.array(y_pred_prob), real_label=np.array(labels))
        write_out(args, "Net", acc, prec, f1, auc, recall)
        # ml model
        for model_name in ["RF", "XGboost", "svm"]:
            model, y_pred_prob = predict_ml(args, data, model_name)
            # # evaluate model
            acc, prec, f1, auc, recall = evaluate(pred_prob=np.array(y_pred_prob), real_label=np.array(labels))
            write_out(args, model_name, acc, prec, f1, auc, recall)




def map_label(df, column, mapping):
    df[column] = df[column].replace(mapping)
    return df

def map_mult_labels(df, columns, labels):
    for column in columns:
        df = map_label(df, labels[column][0], labels[column][1])
    
    return df

#https://stackoverflow.com/questions/26886653/pandas-create-new-column-based-on-values-from-other-columns-apply-a-function-o
def surv_event(row, surv_year):
    if row['Event'] in ["DECEASED", "Recurred/Progressed", "PROGRESSION"]:
        if row['Time'] <= surv_year:
            return 1
        else:
            return 0
    else:
        return 0
    
def do_survival(clin_df, type_surv, labels_mapping, surv_year=120, filename="survival.png"): #type=[OS, DFS, RFS]
    import pandas as pd
    import matplotlib.pyplot as plt

    
    #header = ["ERStatus", "HER2Status", "Pam50Subtype", "Age", "Time", "Event"]
    header = ["Pam50Subtype", "Time", "Event"]
    df_surv = pd.DataFrame(index=clin_df.index, columns=header)
    #df_surv.ERStatus = clin_df.loc[df_surv.index, ['ERStatus']]
    #df_surv.HER2Status = clin_df.loc[df_surv.index, ['HER2Status']]
    df_surv.Pam50Subtype = clin_df.loc[df_surv.index, ['Pam50Subtype']]
    #df_surv.Age = clin_df.loc[df_surv.index, ['Age']]
    if type_surv == "OS":
        df_surv.Time = clin_df.loc[df_surv.index, ['OS_MONTHS']]
        df_surv.Event = clin_df.loc[df_surv.index, ['OS_STATUS']]
    elif type_surv == "DFS":
        df_surv.Time = clin_df.loc[df_surv.index, ['DFS_MONTHS']]
        df_surv.Event = clin_df.loc[df_surv.index, ['DFS_STATUS']]
    elif type_surv == "RFS":
        df_surv.Time = clin_df.loc[df_surv.index, ['RFS_MONTHS']]
        df_surv.Event = clin_df.loc[df_surv.index, ['RFS_STATUS']]    
    
    df_surv["SurvEvent"]=df_surv.apply(lambda row: surv_event(row, surv_year), axis=1)
    df_surv.loc[df_surv['Time']>surv_year, 'SurvTime'] = surv_year
    df_surv.loc[df_surv['Time']<=surv_year, 'SurvTime'] = df_surv['Time']

    df_surv_final = df_surv.drop(['Time', 'Event'], axis=1)
    print (df_surv_final.shape)
    print (sum(df_surv_final.SurvTime.isna()))
    df_surv_final = df_surv_final[~df_surv_final.SurvTime.isna()]
    print (sum(df_surv_final.SurvTime.isna()))
    print (df_surv_final.shape)
    
    from lifelines import CoxPHFitter
    cph = CoxPHFitter()
    cph.fit(df_surv_final, duration_col='SurvTime', event_col='SurvEvent')
    cph.print_summary() 
    
    #cph.plot()
    #cph.plot_covariate_groups('ERStatus', [3,2,1,0])
    
    from lifelines import KaplanMeierFitter
    kmf = KaplanMeierFitter()
    fig = plt.figure(figsize=(10,10))
    ax = plt.subplot(111)
    for name, grouped_df in df_surv_final.groupby('Pam50Subtype'):
        print (name)
        kmf.fit(grouped_df["SurvTime"], grouped_df["SurvEvent"], label=name)
        kmf.plot(ax=ax, ci_show=False, linewidth=4, color=['firebrick', 'hotpink', 'darkblue', 'aqua'][name])
        
    ax.legend(labels_mapping)
    plt.savefig(filename)
    return cph, kmf       


""" Example for biomarker identification
"""
import os
import copy
from feat_importance import cal_feat_imp, summarize_imp_feat

if __name__ == "__main__":
    data_folder = 'BRCA'
    model_folder = os.path.join(data_folder, 'models')
    view_list = [1,2,3]
    if data_folder == 'ROSMAP':
        num_class = 2
    if data_folder == 'BRCA':
        num_class = 5

    featimp_list_list = []
    for rep in range(5):
        featimp_list = cal_feat_imp(data_folder, os.path.join(model_folder, str(rep+1)), 
                                    view_list, num_class)
        featimp_list_list.append(copy.deepcopy(featimp_list))
    summarize_imp_feat(featimp_list_list)
    


import sys
import pandas as pd


def read_(file):
    # read file
    if file.endswith('.csv'):
        df = pd.read_csv(file, index_col=0)
    elif file.endswith('.csv.gz'):
        df = pd.read_csv(file, compression='gzip', index_col=0)
    else:
        print('\n[Error]: The program cannot infer the format of {} . Currently, only the csv format is supported, please ensure that the file name suffix is .csv or .csv.gz.'.format(file))
        sys.exit(0)
    return df


def read_omics(args):
    omics = []
    for file in args.omic_file:
        df = read_(file)
        df = df.fillna(0)  # fill nan with 0
        omics.append(df)
    return omics


def read_label(args):
    file = args.label_file
    df = read_(file)
    df = df.rename(
        columns={df.columns.values[0]: 'label'})
    return df


def read_clin(args):
    file = args.clin_file
    df = None
    if not file is None:
        df = read_(file)
        # fill na
        df = df.fillna(0)
    return df

def process(df_omics, df_label, df_clin):
    # extract patient id
    patients = [df_tmp.index.to_list() for df_tmp in df_omics]
    patients.append(df_label.index.to_list())
    if not df_clin is None:
        patients.append(df_clin.index.to_list())

    # get shared patients between different data
    patients_shared = patients[0]
    for i in range(1, len(patients)):
        patients_shared = list(set(patients_shared).intersection(patients[i]))

    # extract shared patients' data
    for i in range(len(df_omics)):
        df_omics[i] = df_omics[i].loc[patients_shared, :].sort_index()
    df_label = df_label.loc[patients_shared, :].sort_index()
    if not df_clin is None:
        df_clin = df_clin.loc[patients_shared, :].sort_index()
    return df_omics, df_label, df_clin


# api
def read_dataset(args):
    # 1. read raw dataset
    # (1) read omics dataset
    df_omics = read_omics(args)
    # (2) read label
    df_label = read_label(args)
    # (3) read clinical feature
    df_clin = read_clin(args)

    # 2. process
    df_omics, df_label, df_clin = process(df_omics, df_label, df_clin)

    # 3. return clean dataset
    return df_omics, df_label, df_clin



import os, sys, time, torch, gzip
import numpy as np
import pandas as pd
from sklearn.metrics import f1_score, recall_score, roc_auc_score, precision_score, accuracy_score
from captum.attr import IntegratedGradients
from sklearn.impute import KNNImputer


def clin_read_tsi(in_file, out_file, task, threshold=2):
    """To process raw clinical dataset download from linkedomics, which is .tsi format.

    Args:
        in_file (string): The raw clinical dataset file, end with .tsi, a tab delinimated file.
        out_file (string): Output file name. A .csv format file
        task(string): Define task type. Could be one of the following tasks: LST, pan-class
        threshold (int, optional): a threshold year using to split patients into LTS and Non-LTS. Defaults to 2.
    Return:
        return .csv format file. 
    """
    df_clin = pd.read_table(in_file, index_col=0)
    # format the table
    df_clin = df_clin.T
    df_clin = df_clin[~df_clin.overall_survival.isna()]
    df_clin = df_clin[~df_clin.status.isna()]
    df_clin.overall_survival = df_clin.overall_survival.astype('int')
    df_clin.status = df_clin.status.astype('int')
    if task=="LST":
        # label patients
        df_clin.loc[(df_clin.overall_survival < threshold*365) & (df_clin.status == 1), 'label'] = 1
        df_clin.loc[(df_clin.overall_survival >= threshold*365), 'label'] = 0
        df_clin = df_clin[~df_clin.label.isna()]
    else:
        # label patients
        value_map = dict((v, i) for i, v in enumerate(pd.unique(df_clin["histological_type"])))
        print("\nLabels are encoded in to {} categories, encoding dictionary is: {}".format(len(value_map), value_map))
        df_clin["label"] = df_clin["histological_type"]
        df_clin = df_clin.replace({"label":value_map})
    # output: clinical file and label file
    df_clin.to_csv(out_file + "_clinical.csv", index=True)
    label = df_clin[["label"]].astype("int64")
    label.to_csv(out_file + "_label.csv", index=True)


# clinical feature processing
# add mutation info to clinical feature - CGC genes and total mutated gene number
# one hot encode of features
def process_clin(df_clin, df_mut, outfile, task):
    df_mut = df_mut.fillna(0)
    df = df_clin.merge(df_mut, left_index=True, right_index=True, how="left")
    if task == "LST":
        df = df.drop(["overall_survival", "status", "overallsurvival", "label"], axis=1)
        categorical = df.columns[(df.dtypes == "object").values].to_list()
        df = pd.concat([df.drop(categorical, axis=1),
                        pd.get_dummies(df[categorical])],
                       axis=1)
    else:
        df = df.drop(["histological_type", "label", "overallsurvival"], axis=1)
        categorical = df.columns[(df.dtypes == "object").values].to_list()
        df = pd.concat([df.drop(categorical, axis=1),
                        pd.get_dummies(df[categorical])],
                       axis=1)
    df.to_csv(outfile + "_clinical_tmp.csv", index=True)


def check_files(files):
    """To check files.
    files (str or list)
    """
    if isinstance(files, list):
        for f in files:
            if not os.path.exists(f):
                print('[Error] {} not found.'.format(f))
                sys.exit(1)
    elif isinstance(files, str):
        if not os.path.exists(files):
            print('[Error] {} not found.'.format(files))
            sys.exit(1)
    else:
        print('[Error] {} file path is wrong.'.format(files))
        sys.exit(1)


# init log.txt
def init_log(args):
    with open(os.path.join(args.outdir, 'log.txt'), 'a') as file:
        run_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        head_line = "Perform model training at {}".format(run_time)
        file.writelines(head_line + '\n\n')
        if args.FSD:
            if args.clin_file:
                file.writelines("-------Using FSD and clinical file, feature selection method {}, omic group {} clin, and model {}------------\n".format(args.method, args.omic_name, args.model))
            else:
                file.writelines("-------Using FSD, feature selection method {}, omic group {}, and model {}------------\n".format(args.method, args.omic_name, args.model))
        else:
            file.writelines("-------No FSD, using feature selection method {}, omic group {}, and model {}------------\n".format(args.method, args.omic_name, args.model))
    file.close()

# for evaluation
def evaluate(pred_prob, real_label, average="macro"):
    # For evaluating binary classification models
    if pred_prob.shape[1] == 2:
        y_pred = np.argmax(pred_prob, 1)
        prec = precision_score(real_label, y_pred)
        acc = accuracy_score(real_label, y_pred)
        f1 = f1_score(real_label, y_pred)
        recall = recall_score(real_label, y_pred)
        auc = roc_auc_score(real_label, pred_prob[:, 1])
    # For evaluating multiclass models
    else:
        y_pred = np.argmax(pred_prob, 1)
        prec = precision_score(real_label, y_pred, average='macro')
        acc = accuracy_score(real_label, y_pred)
        f1 = f1_score(real_label, y_pred, average=average)
        recall = recall_score(real_label, y_pred, average=average)
        auc = roc_auc_score(real_label, pred_prob, average='macro', multi_class='ovo')
    return acc, prec, f1, auc, recall


# for explain dnn model
def ig(args, model, dataset, feature_names, omic_group, target=1):
    # prepare input data
    input_tensor_list = [data for data, labels in dataset]
    input_tensor = torch.cat(input_tensor_list, 0)
    input_tensor.requires_grad_()

    # instantiation
    ig = IntegratedGradients(model)

    # calculating feature importance using IG
    attr, _ = ig.attribute(input_tensor, return_convergence_delta=True, target=target)
    attr = attr.detach().numpy()
    feat_importance = np.mean(attr, axis=0)

    # result
    df_imp = pd.DataFrame({'Feature': feature_names,
                           'Omic': omic_group,
                           'Target': [target] * len(feature_names),
                           'Importance_value': feat_importance,
                           'Importance_value_abs': abs(feat_importance)
                           })
    df_imp = df_imp.sort_values('Importance_value_abs', ascending=False)

    # output
    if args.FSD:
        if args.clin_file:
            df_imp.to_csv(os.path.join(args.outdir, 'feature_importance_FSD_{}_clin_DNN_{}_target{}.csv'.format(args.method, args.omic_name, target)), index=False)
        else:
            df_imp.to_csv(os.path.join(args.outdir, 'feature_importance_FSD_{}_DNN_{}_target{}.csv'.format(args.method, args.omic_name, target)),index=False)
    else:
        if args.clin_file:
            df_imp.to_csv(os.path.join(args.outdir, 'feature_importance_{}_clin_DNN_{}_target{}.csv'.format(args.method, args.omic_name, target)), index=False)
        else:
            df_imp.to_csv(os.path.join(args.outdir, 'feature_importance_{}_DNN_{}_target{}.csv'.format(args.method, args.omic_name, target)),index=False)

    return df_imp

 # for explain net model
def ig_net(args, model, dataset, feature_names, omic_group, target=1):
    # prepare input data
    input_tensor_dna, input_tensor_rna = [], []
    for data_dna, data_rna, labels in dataset:
        input_tensor_dna.append(data_dna)
        input_tensor_rna.append(data_rna)
    input_tensor_dna = torch.cat(input_tensor_dna, 0).requires_grad_()
    input_tensor_rna = torch.cat(input_tensor_rna, 0).requires_grad_()

    # instantiation
    ig = IntegratedGradients(model)

    # calculating feature importance using IG
    attr, _ = ig.attribute((input_tensor_dna, input_tensor_rna), return_convergence_delta=True, target=target)
    feat_importance = []
    for tensor in attr:
        tensor = tensor.detach().numpy()
        feat_importance.append(np.mean(tensor, axis=0))

    # result
    df_imp = pd.DataFrame({'Feature': feature_names,
                           'Omic': omic_group,
                           'Target': [target] * len(feature_names),
                           'Importance_value': np.concatenate(feat_importance),
                           'Importance_value_abs':  abs(np.concatenate(feat_importance))
                           })
    df_imp = df_imp.sort_values('Importance_value_abs', ascending=False)

    # output
    if args.FSD:
        if args.clin_file:
            df_imp.to_csv(
                os.path.join(args.outdir, 'feature_importance_FSD_{}_clin_Net_{}_target{}.csv'.format(args.method, args.omic_name, target)),
                index=False)
        else:
            df_imp.to_csv(
                os.path.join(args.outdir, 'feature_importance_FSD_{}_Net_{}_target{}.csv'.format(args.method, args.omic_name, target)),
                index=False)
    else:
        if args.clin_file:
            df_imp.to_csv(
                os.path.join(args.outdir, 'feature_importance_{}_clin_Net_{}_target{}.csv'.format(args.method, args.omic_name, target)),
                index=False)
        else:
            df_imp.to_csv(os.path.join(args.outdir, 'feature_importance_{}_Net_{}_target{}.csv'.format(args.method, args.omic_name, target)),
                          index=False)
    return df_imp


def plot_scatter(tsne_comp, samplenames, clin_df, colname, filename): #colname=Pam50Subtype
    import pandas as pd
    import matplotlib.pyplot as plt
    
    import matplotlib.pylab as pylab
    params = {'legend.fontsize': 'x-large',
          'figure.figsize': (15, 5),
         'axes.labelsize': 'large',
         'axes.titlesize':'x-large',
         'xtick.labelsize':'x-large',
         'ytick.labelsize':'x-large',
         'axes.labelweight': 'bold',
             'font.weight' : 'bold',
              'axes.facecolor': 'white',}
    pylab.rcParams.update(params)    
    
    plot_df = pd.DataFrame(data=tsne_comp, 
                          columns = ['comp 1', 'comp 2'], index=samplenames)
    plot_df = pd.concat([plot_df, clin_df[colname]], axis=1)
    print(plot_df.shape)

    targets = [x for x in list(plot_df[colname].unique()) if str(x) != 'nan']
    targets.sort()
    print(targets)
    colors = ['firebrick', 'hotpink', 'darkblue', 'aqua', 'm', 'y', 'b'][0:len(targets)]

    fig = plt.figure(figsize = (10,10))
    ax = fig.add_subplot(1,1,1)
    #ax.set_xlabel('Component 1', fontsize = 15)
    #ax.set_ylabel('Component 2', fontsize = 15)
    
    if colname == "Pam50Subtype":
        ax.set_title("PAM50 Subtypes", fontsize = 20)
    else:
        ax.set_title(colname, fontsize = 20)

    for target, color in zip(targets,colors):
        indicesToKeep = plot_df[colname] == target
        ax.scatter(plot_df.loc[indicesToKeep, 'comp 1']
                , plot_df.loc[indicesToKeep, 'comp 2']
                , c = color
                , s = 40
                , alpha = 1)
    
    ax.legend(targets, fontsize=20)    
    ax.grid()
    plt.savefig(filename)
    
def plot_scatter_kmeans(tsne_comp, samplenames, kmeans_group):
    import pandas as pd
    import matplotlib.pyplot as plt
    
    plot_df = pd.DataFrame(data=tsne_comp, 
                          columns = ['comp 1', 'comp 2'], index=samplenames)
    plot_df['clusters'] = kmeans_group

    colname='clusters'
    targets = [x for x in list(plot_df[colname].unique()) if str(x) != 'nan']
    print(targets)
    colors = ['r', 'hotpink', 'darkblue', 'aqua', 'm', 'y', 'b'][0:len(targets)]

    fig = plt.figure(figsize = (12,12))
    ax = fig.add_subplot(1,1,1)
    #ax.set_xlabel('PC1', fontsize = 15)
    #ax.set_ylabel('PC2', fontsize = 15)
    ax.set_title(colname, fontsize = 20)

    for target, color in zip(targets,colors):
        indicesToKeep = plot_df[colname] == target
        ax.scatter(plot_df.loc[indicesToKeep, 'comp 1']
                , plot_df.loc[indicesToKeep, 'comp 2']
                , c = color
                , s = 10
                , alpha = 0.8)
    
    ax.legend(targets, fontsize=20)    
    ax.grid()
    
def do_TSNE(dat_df, clin_df, column, plot_name):
    from sklearn.manifold import TSNE
    tsne_comp = TSNE().fit_transform(dat_df)
    
    samplenames = clin_df.index.tolist()
    plot_scatter(tsne_comp, samplenames, clin_df, column, plot_name)
    
    return tsne_comp, samplenames

def clustering_metrics(y_true, y_pred):
    from sklearn import metrics
    
    metrics_list = {'ARI': metrics.adjusted_rand_score, 
                    'NMI': metrics.normalized_mutual_info_score, 
                    'FM': metrics.fowlkes_mallows_score
                   }
    
    metrics_dict = {}
    for m, comp_fn in metrics_list.items():
        #print (m, comp_fn(y_true, y_pred))
        metrics_dict[m] = comp_fn(y_true, y_pred)
        
    return metrics_dict 


import torch
import torch.nn as nn
import torch.nn.functional as F


class DeepMOI(nn.Module):
    def __init__(self, in_feat, out_feat):
        super(DeepMOI, self).__init__()
        self.lin1 = nn.Linear(in_feat, 256)
        self.lin2 = nn.Linear(256, 64)
        self.lin3 = nn.Linear(64, out_feat)
        
    def forward(self, x):
        x = self.lin1(x)
        x = torch.relu(x)
        x = F.dropout(x, 0.5)
        
        x = self.lin2(x)
        x = torch.relu(x)
        x = F.dropout(x, 0.5)
        
        x = self.lin3(x)
        
        return x


# class Net(nn.Module):
#     def __init__(self, dim_dna, dim_rna, dim_out):
#         """
#         dim_dna: DNA 组学数据特征维度
#         dim_rna: RNA 组学数据特征维度
#         dim_out: 分类任务的类别数量
#         """
#         super().__init__()
#         # 用于提取DNA组学特征的模块
#         self.dna = nn.Sequential(
#             nn.Linear(dim_dna, 128),
#             nn.BatchNorm1d(num_features=128),
#             nn.Linear(128, dim_dna),
#         )
#         # 用于提取RNA组学特征的模块
#         self.rna = nn.Sequential(
#             nn.Linear(dim_rna, 128),
#             nn.BatchNorm1d(num_features=128),
#             nn.Linear(128, dim_rna),
#         )
#         # 用于分类任务的模块
#         self.mlp = nn.Sequential(
#             nn.Linear(dim_dna + dim_rna, 128),
#             nn.ReLU(),
#             nn.Dropout(p=0.2),
#             nn.BatchNorm1d(num_features=128),
#             nn.Linear(128, dim_out)
#         )
#         self.norm = nn.BatchNorm1d(num_features=dim_dna + dim_rna)

#     def forward(self, data_dna, data_rna):
#         # 提取DNA特征
#         feat_dna = self.dna(data_dna)
#         # 提取RNA特征
#         feat_rna = self.rna(data_rna)
#         # 合并特征
#         h = torch.cat([feat_dna, feat_rna], dim=1)
#         h = self.norm(h)
#         h = h.squeeze(-1)
#         # 分类
#         out = self.mlp(h)
#         return out


class Net(nn.Module):
    def __init__(self, dim_dna, dim_rna, dim_out):
        """
        dim_dna: DNA 组学数据特征维度
        dim_rna: RNA 组学数据特征维度
        dim_out: 分类任务的类别数量
        """
        super().__init__()
        # 用于提取DNA组学特征的模块
        self.dna = nn.Sequential(
            nn.BatchNorm1d(num_features=dim_dna),
            nn.TransformerEncoderLayer(d_model=dim_dna, nhead=1, dim_feedforward=128),
        )
        # 用于提取RNA组学特征的模块
        self.rna = nn.Sequential(
            nn.BatchNorm1d(num_features=dim_rna),
            nn.TransformerEncoderLayer(d_model=dim_rna, nhead=1, dim_feedforward=128),
        )
        # 用于分类任务的模块
        self.mlp = nn.Sequential(
            nn.Linear(dim_dna + dim_rna, 128),
            nn.ReLU(),
            nn.Dropout(p=0.2),
            nn.BatchNorm1d(num_features=128),
            nn.Linear(128, dim_out)
        )
        self.norm = nn.BatchNorm1d(num_features=dim_dna + dim_rna)

    def forward(self, data_dna, data_rna):
        # 提取DNA特征
        feat_dna = self.dna(data_dna)
        # 提取RNA特征
        feat_rna = self.rna(data_rna)
        # 合并特征
        h = torch.cat([feat_dna, feat_rna], dim=1)
        h = self.norm(h)
        h = h.squeeze(-1)
        # 分类
        out = self.mlp(h)
        return out



# Libraries
import torch
import torch.nn as nn
import numpy as np
import pandas as pd
from tensorboardX import SummaryWriter
import torch.nn.functional as F
import pdb
import seaborn as sns
import matplotlib.pyplot as plt
sns.set(style="whitegrid")
from livelossplot import PlotLosses

# Create Fully Conn NN 
class RLModel(nn.Module):
    def __init__(self, input_size, hidden_size, num_classes, n_layers = 3, p = 0.1):
        super(RLModel, self).__init__()
        self.n_layers = n_layers
        self.p = p
        self.fc1 = nn.Linear(input_size, hidden_size) 
        self.activation = nn.ReLU()
        #self.layer = LayerBlock(hidden_size, self.p)
        #self.layers = [LayerBlock(hidden_size , self.p) for i in range(self.n_layers)]
        self.fc2 = nn.Linear(hidden_size, num_classes)
        #self.sigmoid = nn.Sigmoid()
        
    def forward(self, x):
        x = self.fc1(x)
        x = self.activation(x)
        #x = self.layer(x)
        #for i in range(self.n_layers):
        #    x = self.layers[i](x)
        out = self.fc2(x)
        return out
    
def train_RLModel(input_tensor, 
                  target_tensor, 
                  rl_model,
                  rl_model_optimizer,
                #   criterion,
                  phase):
    
    if phase == 'train':
        rl_model.train()
    else:
        rl_model.eval()
        
    # clear gradients in the optimizers
    if phase == 'train':
        rl_model_optimizer.zero_grad()
    
    # Predict label
    prediction = rl_model(input_tensor)
    print(prediction.shape, target_tensor.shape)
    # compute the loss
    loss = torch.nn.functional.binary_cross_entropy_with_logits(prediction, target_tensor)
    
    # # compute the Accuracy
    # _, predicted = torch.max(prediction,1)
    # correct = torch.sum(predicted==target_tensor)
    # total = input_tensor.size(0)
    # accuracy = 100 * correct / total
    
    # if phase == 'train':
    #     # Compute the gradients
    #     loss.backward()
        
    #     # Step the optimizers to update the model weights
    #     rl_model_optimizer.step()
        
    # # return the loss value to track training progress
    # return loss.item(), accuracy.item()

def train_RLModel_Iters(rl_model, 
                        data_tensor_train, 
                        data_tensor_valid,
                        label_tensor_train,
                        label_tensor_valid,
                        num_epochs,
                        print_every_n_batches=100,
                        learning_rate=0.01,
                        phases=['train', 'validation']):
    
    # Live Loss
    liveloss = PlotLosses()
    
    # Keep track of losses
    train_plot_losses = []
    test_plot_losses = []
    train_plot_accuracy = []
    test_plot_accuracy = []
    
    # Initialise RL_Model Optimizer
    rl_model_optimizer = torch.optim.SGD(rl_model.parameters(), lr=learning_rate, momentum=0.9)
    
    # Specify loss function
    criterion = torch.nn.CrossEntropyLoss()
        
    # Cycle through epochs
    for epoch in range(num_epochs):
        logs = {}
        
        for phase in phases:
            print(f'Epoch {epoch +1}/{num_epochs}')
            
            if phase == 'train':
                loss, accuracy = train_RLModel(data_tensor_train, label_tensor_train, rl_model, rl_model_optimizer, criterion, phase)
                train_plot_losses.append(loss)
                train_plot_accuracy.append(accuracy)
            else:
                loss, accuracy = train_RLModel(data_tensor_valid, label_tensor_valid, rl_model, rl_model_optimizer, criterion, phase)
                test_plot_losses.append(loss)
                test_plot_accuracy.append(accuracy)
                
            print(loss)
            #plot_losses.append(loss)
            
            prefix = ''
            if phase == 'validation':
                prefix = 'val_'
                
            logs[prefix + 'log loss'] = loss
            logs[prefix + 'accuracy'] = accuracy
            
            print ('Epoch [{}/{}], Test Loss: {:.4f}' 
                   .format(epoch+1, num_epochs, loss))
            
        liveloss.update(logs)
        liveloss.draw()
        
    # Final accuracy count
    predictions_validation = rl_model(data_tensor_valid)
    _, predicted = torch.max(predictions_validation, 1)
    correct = (predicted==label_tensor_valid).sum().item()
    total = data_tensor_valid.size(0)
    valid_accuracy = 100*correct/total
    print ("Validation Accuracy:", 100*correct/total)
        
    return train_plot_losses, train_plot_accuracy, test_plot_losses, test_plot_accuracy


if __name__ == '__main__':
    X = torch.rand(size=(3, 16, 32))
    Y = np.random.choice(2, size=(3, 16, 1))
    Y = torch.Tensor(Y).to(torch.float)
    rlmodel = RLModel(input_size=32, hidden_size=32, num_classes=1)
    train_RLModel(input_tensor=X, target_tensor=Y, rl_model=rlmodel, 
                  rl_model_optimizer=torch.optim.Adam(rlmodel.parameters()),
                  phase='train'
                  )




from setuptools import setup, find_packages


install_packages = [
    'captum==0.4.1',
    'mygene==3.2.2',
    'openpyxl==3.0.9',
    'packaging==21.3',
    'pandas==1.2.5',
    'pandocfilters==1.5.0',
    'seaborn==0.11.2',
    'torch==1.13.1+cpu',
    'scikit-learn==1.2.2',
    'numpy==1.23.5',
    'matplotlib==3.6.2',
    'xgboost==1.7.4',
    'livelossplot', 
    'tensorboardX',
    'tqdm'
]

setup(
    # 应用名
    name='AttentionMOI',
    # 作者名
    author='Billy',
    # 作者邮箱
    author_email='liangbilin0324@163.com',
    # 版本号
    version='0.0.3',
    # 要求python版本
    python_requires=">=3.9.*",
    # 找到本目录下的所有python包
    packages=find_packages(),
    # 自动安装依赖
    install_requires=install_packages,
    dependency_links=[
        "https://pypi.org/simple/",
        "https://download.pytorch.org/whl/cpu",
        ],
    # 程序网站
    url='https://github.com/BioAI-kits/AttentionMOI',
    # 程序简单描述
    description="A Denoised Multi-omics Integration Framework for Cancer Subtype Classification and Survival Prediction.",
    # 开源许可
    license='Apache License 2.0',
    # 包含的数据
    data_files=['AttentionMOI/example/cnv.csv.gz', 'AttentionMOI/example/met.csv.gz', 'AttentionMOI/example/rna.csv.gz', 'AttentionMOI/example/label.csv'],
    # 命令行
    entry_points={
        'console_scripts': ['moi = moi:moi',
                            ],
    },
)





import sys
import os
import pandas as pd
import numpy as np
from scipy import stats
from sklearn.model_selection import train_test_split
from .preprocess import read_dataset
from tqdm import tqdm
import torch
from .feature_selection import anova, rfe, lasso, pca  # import methods of Feature selection


def distribution_test_binaryclass(data, labels, sample_size=0.3, seed=0):
    # check label0 vs. label1: need (p<0.05)
    label_1 = [True if i == 1 else False for i in labels]
    label_0 = [True if i == 0 else False for i in labels]
    pvalue_01 = stats.kstest(data[label_1], data[label_0]).pvalue

    # check train vs. test for total data: need (p>0.05)
    _, X_smp, _, Y_smp = train_test_split(
        data, labels, test_size=sample_size, random_state=seed)
    label_1 = [True if i == 1 else False for i in Y_smp]
    label_0 = [True if i == 0 else False for i in Y_smp]
    pvalue_02 = stats.kstest(X_smp[label_1], X_smp[label_0]).pvalue
    pvalue_03 = stats.kstest(X_smp, data).pvalue

    return pvalue_01, pvalue_02, pvalue_03


def distribution_test_multiclass(data, labels, sample_size=0.3, seed=0):
    # check label0 vs. label1: need (p<0.05)
    data_i = []
    for i in set(labels):
        idx = [True if n == i else False for n in labels]
        data_i.append(data[idx])

    _, pvalue_01 = stats.f_oneway(*data_i)

    # check train vs. test for total data: need (p>0.05)
    _, X_smp, _, Y_smp = train_test_split(
        data, labels, test_size=sample_size, random_state=seed)
    data_i = []
    for i in set(Y_smp):
        idx = [True if n == i else False for n in Y_smp]
        data_i.append(X_smp[idx])
    _, pvalue_02 = stats.f_oneway(*data_i)

    pvalue_03 = stats.f_oneway(X_smp, data).pvalue

    return pvalue_01, pvalue_02, pvalue_03


def feature_selection_distribution(args):
    # 1. to read input files
    df_omics, df_label, df_clin = read_dataset(args)
    labels = df_label.label.values

    # 2. fix random seed
    np.random.seed(args.seed)
    seeds = np.random.randint(0, 10000, args.iteration)

    # 3. selection: omic-wise selection
    chosen_omics = []  # save feature
    chosen_feat_name = []  # save feature's name
    chosen_omic_group = []  # save feature's omic group

    for omic_num, df_omic in enumerate(df_omics):
        # FSD to remove data noise
        print("Running feature selection for {}\n".format(args.omic_name[omic_num]))
        # use FSD for feature selections
        if args.FSD:
            chosen_feat_id = []  # save feature's id
            omic_data = df_omic.values
            for i in tqdm(range(omic_data.shape[1])):
                data = omic_data[:, i]
                m = 0
                for seed in seeds:
                    # 二分类采用KS检验数据分布
                    if len(set(labels)) == 2:
                        pvalue_01, pvalue_02, pvalue_03 = distribution_test_binaryclass(
                            data=data, labels=labels, seed=seed)
                    # 多分类采用F检验数据分布
                    else:
                        pvalue_01, pvalue_02, pvalue_03 = distribution_test_multiclass(
                            data=data, labels=labels, seed=seed)

                    if pvalue_01 < 0.1 and pvalue_02 < 0.1 and pvalue_03 > 0.05:
                        m += 1
                m = m / len(seeds)
                if m > args.threshold:
                    chosen_feat_id.append(i)

            # 如果没有特征通过FSD筛选
            if len(chosen_feat_id) == 0:
                continue
            df_omic = df_omic.iloc[:, chosen_feat_id]
            print("FSD selected {} features for feature selection.\n".format(df_omic.shape[1]))
            # Feature selection
            if args.method is not None:
                df_omic = feature_selection(args=args, df_omic=df_omic, label=labels)

            # preprocessing result
            chosen_omics.append(df_omic.values)
            chosen_feat_name += df_omic.columns.to_list()  # Filtered feature name
            chosen_omic_group += [args.omic_name[omic_num]] * len(
                df_omic.columns.to_list())  # The omics name corresponding to the filtered feature

            feat = pd.DataFrame({"feat_name": chosen_feat_name,
                                 "omic_group": chosen_omic_group})
            feat.to_csv(os.path.join(args.outdir, 'features_FSD_{}_{}.csv'.format(args.method, args.seed)))
        # No FSD to mitigate noise
        else:
            df_omic = feature_selection(args=args, df_omic=df_omic, label=labels)
            # preprocessing result
            chosen_omics.append(df_omic.values)
            chosen_feat_name += df_omic.columns.to_list()  # Filtered feature name
            chosen_omic_group += [args.omic_name[omic_num]] * len(
                df_omic.columns.to_list())  # The omics name corresponding to the filtered feature

            feat = pd.DataFrame({"feat_name": chosen_feat_name,
                                 "omic_group": chosen_omic_group})
            feat.to_csv(os.path.join(args.outdir, 'features_{}_{}.csv'.format(args.method, args.seed)))

    # 4. add clinical feature
    if df_clin is not None:
        chosen_omics.append(df_clin.values)
        chosen_feat_name += df_clin.columns.to_list()
        chosen_omic_group += ['Clin'] * df_clin.shape[1]

    # 5. merge omic dataset
    data = np.concatenate(chosen_omics, 1)
    data = np.nan_to_num(data)
    data = data.astype('float32')
    if data.shape[1] == 0:
        print(
            'Program terminated! The reason is that no features passed the feature filter. Please check the data or adjust the feature selection parameters.')
        sys.exit(1)

    # 6. return dataset
    return data, chosen_feat_name, chosen_omic_group, labels


def feature_selection(args, df_omic, label):
    """
    select feature according to specific method -- (ANOVA, RFE, LASSO, PCA)

    args: parameters from deepmoi.py.
    df_omic: omic data with DataFrame format.
    label: sample label.

    Return:
        selected omic data with DataFrame format.
    """
    method = args.method
    if method == "ANOVA":
        df_omic = anova(args=args, df_omic=df_omic, label=label)
        print("Using ANOVA for feature selection, {} features were selected for model.".format(df_omic.shape[1]))
        return df_omic
    elif method == "RFE":
        df_omic = rfe(args=args, df_omic=df_omic, label=label)
        print("Using RFE for feature selection, {} features were selected for model.".format(df_omic.shape[1]))
        return df_omic
    elif method == "LASSO":
        df_omic = lasso(df_omic=df_omic, label=label)
        print("Using LASSO for feature selection, {} features were selected for model.".format(df_omic.shape[1]))
        return df_omic
    elif method == "PCA":
        df_omic, feature_imp = pca(args=args, df_omic=df_omic)
        print("Using PCA for feature selection, {} components were used for model.".format(args.num_pc))
        # return df_omic, feature_imp
        return df_omic
    else:
        return df_omic



import os
import torch
import numpy as np
import pandas as pd
import torch.nn as nn
import torch.nn.functional as F
from sklearn import svm
from sklearn.ensemble import RandomForestClassifier
from sklearn.inspection import permutation_importance
from sklearn.model_selection import train_test_split
from torch.utils.data import DataLoader
from xgboost import XGBClassifier
import joblib
from .module import DeepMOI, Net
from .utils import evaluate
from .moanna.model.Moanna import Moanna, Moanna_cls
from .mogonet.main_mogonet import run_mogonet


# builing machine learning models
def ml_models(args, data, chosen_feat_name, chosen_omic_group, labels, model_name):
    """
    args: using args - test_size, seed
    data: combined omics data (format:np.array)
    chosen_feat_name: feature names
    chosen_omic_group: feature of omic group
    labels: labels of classes
    return: print train and test ACC, AUC, Fi_score, Recall, Precision, output feature importance as a .csv file
    """
    X_train, X_test, y_train, y_test = train_test_split(data, labels, test_size=args.test_size, random_state=args.seed)
    if args.model == "RF" or model_name == "RF":
        model = RandomForestClassifier(random_state=args.seed)
        model.fit(X_train, y_train)
        if args.FSD:
            if args.clin_file:
                joblib.dump(model, os.path.join(args.outdir,'model_RF_FSD_{}_{}_clin_{}.joblib'.format(args.method, args.omic_name, args.seed)))
            else:
                joblib.dump(model, os.path.join(args.outdir,'model_RF_FSD_{}_{}_{}.joblib'.format(args.method, args.omic_name, args.seed)))
        else:
            if args.clin_file:
                joblib.dump(model, os.path.join(args.outdir,'model_RF_{}_{}_clin_{}.joblib'.format(args.method, args.omic_name, args.seed)))

            else:
                joblib.dump(model, os.path.join(args.outdir,'model_RF_{}_{}_{}.joblib'.format(args.method, args.omic_name, args.seed)))

    elif args.model == "XGboost" or model_name == "XGboost":
        model = XGBClassifier(random_state=args.seed)
        model.fit(X_train, y_train)
        if args.FSD:
            if args.clin_file:
                joblib.dump(model, os.path.join(args.outdir,'model_XGboost_FSD_{}_{}_clin_{}.joblib'.format(args.method, args.omic_name, args.seed)))
            else:
                joblib.dump(model, os.path.join(args.outdir,'model_XGboost_FSD_{}_{}_{}.joblib'.format(args.method, args.omic_name, args.seed)))
        else:
            if args.clin_file:
                joblib.dump(model, os.path.join(args.outdir,'model_XGboost_{}_{}_clin_{}.joblib'.format(args.method, args.omic_name, args.seed)))

            else:
                joblib.dump(model, os.path.join(args.outdir,'model_XGboost_{}_{}_{}.joblib'.format(args.method, args.omic_name, args.seed)))

    elif args.model == "svm" or model_name == "svm":
        model = svm.SVC(random_state=args.seed, probability=True)
        model.fit(X_train, y_train)
        if args.FSD:
            if args.clin_file:
                joblib.dump(model, os.path.join(args.outdir,'model_svm_FSD_{}_{}_clin_{}.joblib'.format(args.method, args.omic_name, args.seed)))
            else:
                joblib.dump(model, os.path.join(args.outdir,'model_svm_FSD_{}_{}_{}.joblib'.format(args.method, args.omic_name, args.seed)))
        else:
            if args.clin_file:
                joblib.dump(model, os.path.join(args.outdir,'model_svm_{}_{}_clin_{}.joblib'.format(args.method, args.omic_name, args.seed)))

            else:
                joblib.dump(model, os.path.join(args.outdir,'model_svm_{}_{}_{}.joblib'.format(args.method, args.omic_name, args.seed)))

    y_pred_train = model.predict_proba(X_train)
    y_pred_test = model.predict_proba(X_test)
    acc, prec, f1, auc, recall = evaluate(pred_prob=np.array(y_pred_train),
                                          real_label=np.array(y_train))
    print('Train_ACC {:.3f} | Train_AUC {:.3f} | Train_F1_score {:.3f} | Train_Recall {:.3f} | Train_Precision {:.3f}'.format(
                acc, auc, f1, recall, prec))

    acc, prec, f1, auc, recall = evaluate(pred_prob=np.array(y_pred_test),
                                          real_label=np.array(y_test))
    test_evaluate = 'Test_ACC  {:.3f} | Test_AUC  {:.3f} | Test_F1_score  {:.3f} | Test_Recall  {:.3f} | Test_Precision  {:.3f}\n'.format(
                acc, auc, f1, recall, prec)
    print(test_evaluate)
    with open(os.path.join(args.outdir, 'log.txt'), 'a') as file:
        file.writelines("------Running {} model------\n".format(model_name))
        file.writelines(test_evaluate + '\n')
    file.close()

    # write output for model evaluation
    # model, feature selection, accuracy, precision, f1_score, AUC, recall
    with open(os.path.join(args.outdir, 'evaluation.txt'), 'a') as txt:
        if args.FSD:
            if args.clin_file:
                txt.writelines(f"{model_name}\tFSD_{args.method}\t{args.omic_name} clin\t{acc}\t{prec}\t{f1}\t{auc}\t{recall}\n")
            else:
                txt.writelines(f"{model_name}\tFSD_{args.method}\t{args.omic_name}\t{acc}\t{prec}\t{f1}\t{auc}\t{recall}\n")
        else:
            if args.clin_file:
                txt.writelines(f"{model_name}\t{args.method}\t{args.omic_name} clin\t{acc}\t{prec}\t{f1}\t{auc}\t{recall}\n")
            else:
                txt.writelines(
                    f"{model_name}\t{args.method}\t{args.omic_name}\t{acc}\t{prec}\t{f1}\t{auc}\t{recall}\n")
    txt.close()


# using DNN model
def train(args, data, labels):
    dataset = []
    for i in range(len(labels)):
        dataset.append([torch.tensor(data[i], dtype=torch.float),
                        torch.tensor(labels[i], dtype=torch.long)])

    # in_dim, out_dim
    dim_out = len(set(labels))
    in_dim = data.shape[1]
    model = DeepMOI(in_dim, dim_out)
    optimizer = torch.optim.Adam(model.parameters(),
                                 lr=args.lr, weight_decay=args.weight_decay)
    dataset_train, dataset_test = train_test_split(
        dataset, test_size=args.test_size, random_state=args.seed)
    train_loader = DataLoader(dataset_train, batch_size=args.batch)
    test_loader = DataLoader(dataset_test, batch_size=args.batch, shuffle=True, drop_last=True)

    for epoch in range(args.epoch):
        # 1) training model
        model.train()
        loss_epoch = []
        y_pred_probs, real_labels = [], []
        for data, labels in train_loader:
            out = model(data)
            out = out.squeeze(-1)
            # loss
            loss = nn.CrossEntropyLoss()(out, labels)
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            loss_epoch.append(loss.item())
            # prediction
            y_pred_prob = F.softmax(out).detach().tolist()
            y_pred_probs.append(y_pred_prob)
            real_labels.append(labels)
        loss_epoch = np.mean(loss_epoch)
        y_pred_probs = np.concatenate(y_pred_probs)
        real_labels = np.concatenate(real_labels)
        acc, prec, f1, auc, recall = evaluate(y_pred_probs, real_labels)
        log_train = 'Epoch {:2d} | Train Loss {:.10f} | Train_ACC {:.3f} | Train_AUC {:.3f} | Train_F1_score {:.3f} | Train_Recall {:.3f} | Train_Precision {:.3f}'.format(
            epoch, loss_epoch, acc, auc, f1, recall, prec)
        with open(os.path.join(args.outdir, 'log.txt'), 'a') as file:
            file.writelines(log_train + '\n')

        with torch.no_grad():
            # test metrics
            loss_epoch = []
            y_pred_probs, real_labels = [], []
            for data, labels in test_loader:
                out = model(data)
                out = out.squeeze(-1)
                # loss
                loss = nn.CrossEntropyLoss()(out, labels)
                optimizer.zero_grad()
                loss_epoch.append(loss.item())
                # predict labels
                y_pred_prob = F.softmax(out).detach().tolist()
                y_pred_probs.append(y_pred_prob)
                real_labels.append(labels)

            loss_epoch = np.mean(loss_epoch)
            y_pred_probs = np.concatenate(y_pred_probs)
            real_labels = np.concatenate(real_labels)
            acc, prec, f1, auc, recall = evaluate(y_pred_probs, real_labels)
            log_test = 'Epoch {:2d} | Test Loss {:.10f} | Test_ACC {:.3f} | Test_AUC {:.3f} | Test_F1_score {:.3f} | Test_Recall {:.3f} | Test_Precision {:.3f}'.format(
                epoch, loss_epoch, acc, auc, f1, recall, prec)
            print(log_test)
            with open(os.path.join(args.outdir, 'log.txt'), 'a') as file:
                file.writelines(log_test + '\n')
    with open(os.path.join(args.outdir, 'evaluation.txt'), 'a') as txt:
        if args.FSD:
            if args.clin_file:
                txt.writelines(f"DNN\tFSD_{args.method}\t{args.omic_name} clin\t{acc}\t{prec}\t{f1}\t{auc}\t{recall}\n")
                # to save model
                torch.save(model, os.path.join(args.outdir,
                                               'model_DNN_FSD_{}_{}_clin_{}.pt'.format(args.method, args.omic_name, args.seed)))
            else:
                txt.writelines(f"DNN\tFSD_{args.method}\t{args.omic_name}\t{acc}\t{prec}\t{f1}\t{auc}\t{recall}\n")
                # to save model
                torch.save(model, os.path.join(args.outdir,
                                               'model_DNN_FSD_{}_{}_{}.pt'.format(args.method, args.omic_name, args.seed)))
        else:
            if args.clin_file:
                txt.writelines(f"DNN\t{args.method}\t{args.omic_name} clin\t{acc}\t{prec}\t{f1}\t{auc}\t{recall}\n")
                # to save model
                torch.save(model, os.path.join(args.outdir,
                                               'model_DNN_{}_{}_clin_{}.pt'.format(args.method, args.omic_name, args.seed)))
            else:
                txt.writelines(f"DNN\t{args.method}\t{args.omic_name}\t{acc}\t{prec}\t{f1}\t{auc}\t{recall}\n")
                # to save model
                torch.save(model, os.path.join(args.outdir,
                                               'model_DNN_{}_{}_{}.pt'.format(args.method, args.omic_name, args.seed)))
    txt.close()

    return model, test_loader


def train_net(args, data, chosen_omic_group, labels):
    idx_rna = []
    for i, v in enumerate(chosen_omic_group):
        if v == 'rna':
            idx_rna.append(i)

    idx_dna = []
    for i, v in enumerate(chosen_omic_group):
        if v == 'cnv' or v == 'met':
            idx_dna.append(i)

    data_dna = data[:, idx_dna]
    data_rna = data[:, idx_rna]
    dataset = []
    for i in range(len(labels)):
        dataset.append([torch.tensor(data_dna[i], dtype=torch.float),
                        torch.tensor(data_rna[i], dtype=torch.float),
                        torch.tensor(labels[i], dtype=torch.long)])

    # dim_dna, dim_rna, dim_out
    dim_dna = data_dna.shape[1]
    dim_rna = data_rna.shape[1]
    dim_out = len(set(labels))
    model = Net(dim_dna, dim_rna, dim_out)
    optimizer = torch.optim.Adam(model.parameters(),
                                 lr=args.lr,
                                 weight_decay=args.weight_decay)
    dataset_train, dataset_test = train_test_split(
        dataset, test_size=args.test_size, random_state=args.seed)
    train_loader = DataLoader(dataset_train, batch_size=args.batch)
    test_loader = DataLoader(dataset_test, batch_size=args.batch, shuffle=True, drop_last=True)

    with open(os.path.join(args.outdir, 'log.txt'), 'a') as file:
        file.writelines("------Running Net model------" + '\n')

    for epoch in range(args.epoch):
        # 1) training model
        model.train()
        loss_epoch = []
        y_pred_probs, real_labels = [], []
        for data_dna, data_rna, labels in train_loader:
            out = model(data_dna, data_rna)
            out = out.squeeze(-1)
            # loss
            loss = nn.CrossEntropyLoss()(out, labels)
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            loss_epoch.append(loss.item())
            # prediction
            y_pred_prob = F.softmax(out).detach().tolist()
            y_pred_probs.append(y_pred_prob)
            real_labels.append(labels)
        loss_epoch = np.mean(loss_epoch)
        y_pred_probs = np.concatenate(y_pred_probs)
        real_labels = np.concatenate(real_labels)
        acc, prec, f1, auc, recall = evaluate(y_pred_probs, real_labels)
        log_train = 'Epoch {:2d} | Train Loss {:.10f} | Train_ACC {:.3f} | Train_AUC {:.3f} | Train_F1_score {:.3f} | Train_Recall {:.3f} | Train_Precision {:.3f}'.format(
            epoch, loss_epoch, acc, auc, f1, recall, prec)
        print(log_train)
        with open(os.path.join(args.outdir, 'log.txt'), 'a') as file:
            file.writelines(log_train + '\n')

        with torch.no_grad():
            # test metrics
            loss_epoch = []
            y_pred_probs, real_labels = [], []
            for data_dna, data_rna, labels in test_loader:
                out = model(data_dna, data_rna)
                out = out.squeeze(-1)
                # loss
                loss = nn.CrossEntropyLoss()(out, labels)
                optimizer.zero_grad()
                loss_epoch.append(loss.item())
                # predict labels
                y_pred_prob = F.softmax(out).detach().tolist()
                y_pred_probs.append(y_pred_prob)
                real_labels.append(labels)

            loss_epoch = np.mean(loss_epoch)
            y_pred_probs = np.concatenate(y_pred_probs)
            real_labels = np.concatenate(real_labels)
            acc, prec, f1, auc, recall = evaluate(y_pred_probs, real_labels)
            log_test = 'Epoch {:2d} | Test Loss {:.10f} | Test_ACC {:.3f} | Test_AUC {:.3f} | Test_F1_score {:.3f} | Test_Recall {:.3f} | Test_Precision {:.3f}'.format(
                epoch, loss_epoch, acc, auc, f1, recall, prec)
            print(log_test)

            # to write log info
            with open(os.path.join(args.outdir, 'log.txt'), 'a') as file:
                file.writelines(log_test + '\n')

    with open(os.path.join(args.outdir, 'evaluation.txt'), 'a') as txt:
        if args.FSD:
            if args.clin_file:
                txt.writelines(f"Net\tFSD_{args.method}\t{args.omic_name} clin\t{acc}\t{prec}\t{f1}\t{auc}\t{recall}\n")
                # to save model
                torch.save(model, os.path.join(args.outdir,
                                               'model_Net_FSD_{}_{}_clin_{}.pt'.format(args.method, args.omic_name, args.seed)))
            else:
                txt.writelines(f"Net\tFSD_{args.method}\t{args.omic_name}\t{acc}\t{prec}\t{f1}\t{auc}\t{recall}\n")
                # to save model
                torch.save(model, os.path.join(args.outdir,
                                               'model_Net_FSD_{}_{}_{}.pt'.format(args.method, args.omic_name, args.seed)))
        else:
            if args.clin_file:
                txt.writelines(f"Net\t{args.method}\t{args.omic_name} clin\t{acc}\t{prec}\t{f1}\t{auc}\t{recall}\n")
                # to save model
                torch.save(model, os.path.join(args.outdir,
                                               'model_Net_{}_{}_clin_{}.pt'.format(args.method, args.omic_name, args.seed)))
            else:
                txt.writelines(f"Net\t{args.method}\t{args.omic_name}\t{acc}\t{prec}\t{f1}\t{auc}\t{recall}\n")
                # to save model
                torch.save(model, os.path.join(args.outdir,
                                               'model_Net_{}_{}_{}.pt'.format(args.method, args.omic_name, args.seed)))
    txt.close()


    return model, test_loader



def evaluation(model, dataset):
    """
    To evaluate model performance for training dataset and testing dataset.

    model: taining model
    dataset: training dataset or testing dataset.
    """
    y_pred_probs, real_labels = [], []
    for d in dataset:
        out = model(d[0])
        y_pred_prob = F.softmax(out).detach().tolist()
        y_pred_probs.append(y_pred_prob)
        real_labels.append(d[1].tolist())
    acc, prec, f1, auc, recall = evaluate(pred_prob=np.array(y_pred_probs),
                                          real_label=np.array(real_labels))
    return acc, prec, f1, auc, recall


# using Moanna model
def train_moanna(args, data, labels):    
     # Pre-defined hyperparameters
    params = {}
    params["input_size"] = data.shape[1] # Number of features
    params["n_layers"] = 1
    params["encoded_size"] = 64 # Number of Autoencoders neurons
    params["hidden_size"] = params["encoded_size"] * (2 ** (params["n_layers"] + 1))
    params["drop_prob"] = 0.5
    params["fnn_hidden_size"] = 40
    params["fnn_number_layers"] = 2
    params["num_classes"] = len(set(labels))
    params["fnn_num_epoch"] = 100
    params["fnn_learning_rate"] = 0.05

    # Setup model
    model = Moanna(
        params["input_size"], 
        params["hidden_size"], 
        params["encoded_size"], 
        params["n_layers"], 
        params["drop_prob"], 
        params["fnn_hidden_size"], 
        params["num_classes"], 
        params["fnn_number_layers"], 
        0.1,
    ) 
    
    model_cls = Moanna_cls(
        params["input_size"], 
        params["hidden_size"], 
        params["encoded_size"], 
        params["n_layers"], 
        params["drop_prob"], 
        params["fnn_hidden_size"], 
        params["num_classes"], 
        params["fnn_number_layers"], 
        0.1,
    ) 
    
    
    dataset = []
    for i in range(len(labels)):
        dataset.append([torch.tensor(data[i], dtype=torch.float),
                        torch.tensor(labels[i], dtype=torch.long)])

    # in_dim, out_dim
    dim_out = len(set(labels))
    in_dim = data.shape[1]
    optimizer = torch.optim.Adam(model.parameters(),
                                 lr=args.lr, weight_decay=args.weight_decay)
    dataset_train, dataset_test = train_test_split(
        dataset, test_size=args.test_size, random_state=args.seed)
    train_loader = DataLoader(dataset_train, batch_size=args.batch)
    test_loader = DataLoader(dataset_test, batch_size=args.batch, shuffle=True, drop_last=True)

    # 先train 编码器和解码器
    criterion1 = torch.nn.MSELoss(reduce=True)
    for _ in range(50):
        model.train()
        for data, labels in train_loader:
            encoded, decoded = model(data)
            loss = criterion1(decoded, data)
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
    
    for epoch in range(args.epoch):
        # 1) training model
        model_cls.train()
        loss_epoch = []
        y_pred_probs, real_labels = [], []
        for data, labels in train_loader:
            encoded, decoded = model(data)
            out = model_cls(encoded)
    #         out = out.squeeze(-1)
    #         # loss
            loss = nn.CrossEntropyLoss()(out, labels)
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            loss_epoch.append(loss.item())
            # prediction
            y_pred_prob = F.softmax(out).detach().tolist()
            y_pred_probs.append(y_pred_prob)
            real_labels.append(labels)
        loss_epoch = np.mean(loss_epoch)
        y_pred_probs = np.concatenate(y_pred_probs)
        real_labels = np.concatenate(real_labels)
        acc, prec, f1, auc, recall = evaluate(y_pred_probs, real_labels)
        log_train = 'Epoch {:2d} | Train Loss {:.10f} | Train_ACC {:.3f} | Train_AUC {:.3f} | Train_F1_score {:.3f} | Train_Recall {:.3f} | Train_Precision {:.3f}'.format(
            epoch, loss_epoch, acc, auc, f1, recall, prec)
        print(log_train)
        with open(os.path.join(args.outdir, 'log.txt'), 'a') as file:
            file.writelines(log_train + '\n')

        with torch.no_grad():
            # test metrics
            loss_epoch = []
            y_pred_probs, real_labels = [], []
            for data, labels in test_loader:
                encoded, decoded = model(data)
                out = model_cls(encoded)
                # out = out.squeeze(-1)
                # loss
                loss = nn.CrossEntropyLoss()(out, labels)
                optimizer.zero_grad()
                loss_epoch.append(loss.item())
                # predict labels
                y_pred_prob = F.softmax(out).detach().tolist()
                y_pred_probs.append(y_pred_prob)
                real_labels.append(labels)

            loss_epoch = np.mean(loss_epoch)
            y_pred_probs = np.concatenate(y_pred_probs)
            real_labels = np.concatenate(real_labels)
            acc, prec, f1, auc, recall = evaluate(y_pred_probs, real_labels)
            log_test = 'Epoch {:2d} | Test Loss {:.10f} | Test_ACC {:.3f} | Test_AUC {:.3f} | Test_F1_score {:.3f} | Test_Recall {:.3f} | Test_Precision {:.3f}'.format(
                epoch, loss_epoch, acc, auc, f1, recall, prec)
            print(log_test)
            with open(os.path.join(args.outdir, 'log.txt'), 'a') as file:
                file.writelines(log_test + '\n')
    with open(os.path.join(args.outdir, 'evaluation.txt'), 'a') as txt:
        if args.FSD:
            if args.clin_file:
                txt.writelines(f"DNN\tFSD_{args.method}\t{args.omic_name} clin\t{acc}\t{prec}\t{f1}\t{auc}\t{recall}\n")
                # to save model
                torch.save(model, os.path.join(args.outdir,
                                               'model_DNN_FSD_{}_{}_clin_{}.pt'.format(args.method, args.omic_name, args.seed)))
            else:
                txt.writelines(f"DNN\tFSD_{args.method}\t{args.omic_name}\t{acc}\t{prec}\t{f1}\t{auc}\t{recall}\n")
                # to save model
                torch.save(model, os.path.join(args.outdir,
                                               'model_DNN_FSD_{}_{}_{}.pt'.format(args.method, args.omic_name, args.seed)))
        else:
            if args.clin_file:
                txt.writelines(f"DNN\t{args.method}\t{args.omic_name} clin\t{acc}\t{prec}\t{f1}\t{auc}\t{recall}\n")
                # to save model
                torch.save(model, os.path.join(args.outdir,
                                               'model_DNN_{}_{}_clin_{}.pt'.format(args.method, args.omic_name, args.seed)))
            else:
                txt.writelines(f"DNN\t{args.method}\t{args.omic_name}\t{acc}\t{prec}\t{f1}\t{auc}\t{recall}\n")
                # to save model
                torch.save(model, os.path.join(args.outdir,
                                               'model_DNN_{}_{}_{}.pt'.format(args.method, args.omic_name, args.seed)))
    txt.close()


# using MOGONET model
def train_mogonet(args, data, labels):
    dataset = []
    for i in range(len(labels)):
        dataset.append([torch.tensor(data[i], dtype=torch.float),
                        torch.tensor(labels[i], dtype=torch.long)])

    dataset_train, dataset_test = train_test_split(
        dataset, test_size=args.test_size, random_state=args.seed)
    
    os.makedirs('tmp', exist_ok=True)
    dats, labs = [], []
    for dat, label in dataset_train:
        dats.append(dat.numpy())
        labs.append(label.numpy())
    a = np.vstack(dats)
    b = np.vstack(labs)
    np.savetxt('tmp/1_tr.csv', a, delimiter=',')
    np.savetxt('tmp/labels_tr.csv', b, delimiter=',')
    
    dats, labs = [], []
    for dat, label in dataset_test:
        dats.append(dat.numpy())
        labs.append(label.numpy())
    a = np.vstack(dats)
    b = np.vstack(labs)
    np.savetxt('tmp/1_te.csv', a, delimiter=',')
    np.savetxt('tmp/labels_te.csv', b, delimiter=',')
    
    
    logs = run_mogonet(num_epoch=args.epoch)
    with open(os.path.join(args.outdir, 'log.txt'), 'a') as file:
        for log in logs:
            file.writelines(log + '\n')
    


from .Autoencoder import *
from .ClassificationModel import *

class Moanna(nn.Module):
    def __init__(self, input_shape, hidden_size, encoded_size, n_layers, drop_prob, fnn_hidden_size, num_classes, fnn_n_layers, fnn_p):
        super(Moanna, self).__init__()
        self.encoder = Encoder(input_shape, hidden_size, encoded_size, n_layers, drop_prob)
        self.decoder = Decoder(input_shape, hidden_size, encoded_size, n_layers, drop_prob)
        self.classifier = RLModel(encoded_size, fnn_hidden_size, num_classes, fnn_n_layers, fnn_p)
        
    def forward(self, x):
        outs = [] 
        encoded = self.encoder(x)
        decoded = self.decoder(encoded)
        # out = self.classifier(encoded)
        
        return encoded, decoded
    
class Moanna_cls(nn.Module):
    def __init__(self, input_shape, hidden_size, encoded_size, n_layers, drop_prob, fnn_hidden_size, num_classes, fnn_n_layers, fnn_p):
        super(Moanna_cls, self).__init__()
        self.encoder = Encoder(input_shape, hidden_size, encoded_size, n_layers, drop_prob)
        self.decoder = Decoder(input_shape, hidden_size, encoded_size, n_layers, drop_prob)
        self.classifier = RLModel(encoded_size, fnn_hidden_size, num_classes, fnn_n_layers, fnn_p)
        
    def forward(self, encoded):
        # outs = [] 
        # encoded = self.encoder(x)
        # decoded = self.decoder(encoded)
        out = self.classifier(encoded)
        
        return out


import os
import copy
import numpy as np
import pandas as pd
import torch

from sklearn.metrics import f1_score
from utils import load_model_dict
from models import init_model_dict
from train_test import prepare_trte_data, gen_trte_adj_mat, test_epoch

cuda = True if torch.cuda.is_available() else False

def cal_feat_imp(data_folder, model_folder, view_list, num_class):
    num_view = len(view_list)
    dim_hvcdn = pow(num_class,num_view)
    if data_folder == 'ROSMAP':
        adj_parameter = 2
        dim_he_list = [200,200,100]
    if data_folder == 'BRCA':
        adj_parameter = 10
        dim_he_list = [400,400,200]
    data_tr_list, data_trte_list, trte_idx, labels_trte = prepare_trte_data(data_folder, view_list)
    adj_tr_list, adj_te_list = gen_trte_adj_mat(data_tr_list, data_trte_list, trte_idx, adj_parameter)
    featname_list = []
    for v in view_list:
        df = pd.read_csv(os.path.join(data_folder, str(v)+"_featname.csv"), header=None)
        featname_list.append(df.values.flatten())
    
    dim_list = [x.shape[1] for x in data_tr_list]
    model_dict = init_model_dict(num_view, num_class, dim_list, dim_he_list, dim_hvcdn)
    for m in model_dict:
        if cuda:
            model_dict[m].cuda()
    model_dict = load_model_dict(model_folder, model_dict)
    te_prob = test_epoch(data_trte_list, adj_te_list, trte_idx["te"], model_dict)
    if num_class == 2:
        f1 = f1_score(labels_trte[trte_idx["te"]], te_prob.argmax(1))
    else:
        f1 = f1_score(labels_trte[trte_idx["te"]], te_prob.argmax(1), average='macro')
    
    feat_imp_list = []
    for i in range(len(featname_list)):
        feat_imp = {"feat_name":featname_list[i]}
        feat_imp['imp'] = np.zeros(dim_list[i])
        for j in range(dim_list[i]):
            feat_tr = data_tr_list[i][:,j].clone()
            feat_trte = data_trte_list[i][:,j].clone()
            data_tr_list[i][:,j] = 0
            data_trte_list[i][:,j] = 0
            adj_tr_list, adj_te_list = gen_trte_adj_mat(data_tr_list, data_trte_list, trte_idx, adj_parameter)
            te_prob = test_epoch(data_trte_list, adj_te_list, trte_idx["te"], model_dict)
            if num_class == 2:
                f1_tmp = f1_score(labels_trte[trte_idx["te"]], te_prob.argmax(1))
            else:
                f1_tmp = f1_score(labels_trte[trte_idx["te"]], te_prob.argmax(1), average='macro')
            feat_imp['imp'][j] = (f1-f1_tmp)*dim_list[i]
            
            data_tr_list[i][:,j] = feat_tr.clone()
            data_trte_list[i][:,j] = feat_trte.clone()
        feat_imp_list.append(pd.DataFrame(data=feat_imp))
    
    return feat_imp_list


def summarize_imp_feat(featimp_list_list, topn=30):
    num_rep = len(featimp_list_list)
    num_view = len(featimp_list_list[0])
    df_tmp_list = []
    for v in range(num_view):
        df_tmp = copy.deepcopy(featimp_list_list[0][v])
        df_tmp['omics'] = np.ones(df_tmp.shape[0], dtype=int)*v
        df_tmp_list.append(df_tmp.copy(deep=True))
    df_featimp = pd.concat(df_tmp_list).copy(deep=True)
    for r in range(1,num_rep):
        for v in range(num_view):
            df_tmp = copy.deepcopy(featimp_list_list[r][v])
            df_tmp['omics'] = np.ones(df_tmp.shape[0], dtype=int)*v
            df_featimp = df_featimp.append(df_tmp.copy(deep=True), ignore_index=True) 
    df_featimp_top = df_featimp.groupby(['feat_name', 'omics'])['imp'].sum()
    df_featimp_top = df_featimp_top.reset_index()
    df_featimp_top = df_featimp_top.sort_values(by='imp',ascending=False)
    df_featimp_top = df_featimp_top.iloc[:topn]
    print('{:}\t{:}'.format('Rank','Feature name'))
    for i in range(len(df_featimp_top)):
        print('{:}\t{:}'.format(i+1,df_featimp_top.iloc[i]['feat_name']))